#!/usr/bin/env python

# basics
import glob, json, os, pprint, sys
import numpy as np
from multiprocessing import Pool

# pyetc tools  (hopefully not too many here)
import pysynphot
from pysynphot import spparser as SP
import synphot as syn
from astropy import units as u
from astropy import constants as const

# pandeia tools
from pandeia.engine.calc_utils import build_default_calc
from pandeia.engine.perform_calculation import perform_calculation
from pandeia.engine.io_utils import NumPyArangeEncoder
from pandeia.engine.instrument_factory import InstrumentFactory
from pandeia_test import utils, data_util # for rglob - maybe eventually move that to pandeia


# doc
"""
The general idea here is to take in an HST ETC (HETC) engine test file i.e. a
*pyetc* engine input file (*.peng), and convert it to *pandeia* engine inputs
outputting a '*.jeng' file.

We are starting with:
- HST inputs → JWST inputs
- JWST outputs → HST outputs (later)
- assumptions at start:
  - turn the sky off
  - simmode = SNR
"""

DATA_DIR = os.environ['pandeia_refdata']

# data drive this whole thing
HST_INSTS = ['acs','cos','stis','wfc3ir','wfc3uvis']
HST_INSTS_NEW_NAMES = ['wfc3']
HST_DETECTORS = ['ccd','ir','fuv','nuv','fuvmama','nuvmama','sbc','uvis1','uvis2','wfc', 'hrc']
HST_MODES = ['imaging','spectroscopic', 'spectroscopicacq', 'targetacquisition', 'scimaging', 'scspectroscopic', 'rampfilter']
HST_MIRRORS = ['mirrora','mirrorb']

MODE_NAME_DICT = {'spectroscopic': 'spec',
                  'spectroscopicacq': 'specacq',
                  'scspectroscopic': 'scan_spec',
                  'scimaging': 'scan_imaging',
                  'targetacquisition': 'imagingacq'}

DETECTOR_INSTRUMENT_DICT = {'wfc': 'acs',
                            'hrc': 'acs',
                            'sbc': 'acs',
                            'fuv': 'cos',
                            'nuv': 'cos',
                            'ccd': 'stis',
                            'fuvmama': 'stis',
                            'nuvmama': 'stis',
                            'ir': 'wfc3',
                            'uvis1': 'wfc3',
                            'uvis2': 'wfc3'}

APERTURE_DICT = {'psa':
                     {'fuv': 'cosfuv',
                      'nuv': 'cosnuv'},
                 'boa':
                     {'fuv': 'cosfuv',
                      'nuv': 'cosnuv'},
                 'wfc':
                     {'wfc': 'wfc1'}
                 }

# Dictionaries to convert the listed flux/wavelength units in the peng file to their corresponding astropy/pysynphot units.
FLUX_UNIT_DICT = {'photlam':syn.units.PHOTLAM,
                  'photnu':syn.units.PHOTNU,
                  'flam':syn.units.FLAM,
                  'fnu':syn.units.FNU,
                  'jy':u.Jy,
                  'stmag':u.STmag,
                  'abmag':u.ABmag}

WAVE_UNIT_DICT = {'angstroms':u.angstrom,
                  'nm':u.nm,
                  'm':u.m,
                  'micron':u.micron,
                  'microns':u.micron
                  }

SOURCEVAL_FUNC_DICT = {"spec":"sed",
                       "z":"redshift",
                       "icat":"sed",
                       "unit":"sed",
                       "pl":"sed",
                       "bb":"sed",
                       "em":"lines",
                       "rn":"normalization",
                       "ebmvx":"extinction"}

PWEB_FILTER_DICT = {"Johnson/V":"johnson,v", 
                    "Johnson/U":"johnson,u",
                    "Johnson/B":"johnson,b", 
                    "Johnson/R":"johnson,r", 
                    "Johnson/I":"johnson,i",
                    "Johnson/J":"johnson,j",
                    "Johnson/H":"johnson,h",
                    "Johnson/K":"johnson,k",
                    "Cousins/R":"cousins,r",
                    "Cousins/I":"cousins,i", 
                    "Bessell/J":"bessell,j", 
                    "Bessell/H":"bessell,h", 
                    "Bessell/K":"bessell,k",
                    "Sloan/U":"sdss,u",
                    "Sloan/G":"sdss,u",
                    "Sloan/R":"sdss,u",
                    "Sloan/I":"sdss,u",
                    "Sloan/Z":"sdss,u",
                    "Galex/FUV":"galex,fuv",
                    "Galex/NUV":"galex,nuv",
                    "NICMOS/F110W":"nicmos,2,f110w",
                    "NICMOS/F160W":"nicmos,2,f160w",
                    "ACS/F435W":"acs,wfc1,f435w",
                    "ACS/F475W":"acs,wfc1,f475w",
                    "ACS/F555W":"acs,wfc1,f555w",
                    "ACS/F606W":"acs,wfc1,f606w",
                    "ACS/F625W":"acs,wfc1,f625w",
                    "ACS/F775W":"acs,wfc1,f775w",
                    "ACS/F814W":"acs,wfc1,f814w",
                    "ACS/F850LP":"acs,wfc1,f850lp",
                    "WFC3/UVIS/F218W":"wfc3,uvis1,f218w",
                    "WFC3/UVIS/F200LP":"wfc3,uvis1,f200lp",
                    "WFC3/UVIS/F225W":"wfc3,uvis1,f225w",
                    "WFC3/UVIS/F275W":"wfc3,uvis1,f275w",
                    "WFC3/UVIS/F300X":"wfc3,uvis1,f300x",
                    "WFC3/UVIS/F336W":"wfc3,uvis1,f336w",
                    "WFC3/UVIS/F350LP":"wfc3,uvis1,f350lp",
                    "WFC3/UVIS/F390W":"wfc3,uvis1,f390w",
                    "WFC3/UVIS/F438W":"wfc3,uvis1,f438w",
                    "WFC3/UVIS/F475X":"wfc3,uvis1,f475x",
                    "WFC3/UVIS/F600LP":"wfc3,uvis1,f600lp",
                    "WFC3/IR/F098M": "wfc3,ir,f098m",
                    "WFC3/IR/F105W": "wfc3,ir,f105w",
                    "WFC3/IR/F110W": "wfc3,ir,f110w",
                    "WFC3/IR/F125W": "wfc3,ir,f125w",
                    "WFC3/IR/F140W": "wfc3,ir,f140w",
                    "WFC3/IR/F160W": "wfc3,ir,f160w"}


PWEB_NORMUNIT_DICT = {"angstroms": "angstrom",
                      "nanometers": "nm",
                      "microns": "micron"
                        }

UNSUPPORTED_SED_TYPES = []

EXTINCTION_DICT = { "mwavg": "mw_rv_31",
                    "mwrv21": "mw_rv_31",
                    "mwrv4": "mw_rv_40",
                    "mwdense": "mw_rv_55",
                    "lmcavg": "lmc_avg",
                    "lmc30dor": "lmc_2",
                    "smcbar": "smc_bar",
                    "xgalsb": "hd210121"}

K93_MODEL_DICT = {  "44500": "o5v",
                    "38000": "o7v",
                    "33000": "o9v",
                    "30000": "b0v",
                    "25400": "b1v",
                    "18700": "b3v",
                    "15400": "b5v",
                    "11900": "b8v",
                    "9230": "a1v",
                    "8720": "a3v",
                    "8200": "a5v",
                    "7700": "f0i",
                    "7200": "f0v",
                    "6890": "f2v",
                    "6440": "f5v",
                    "6200": "f8v",
                    "5860": "g2v",
                    "5770": "g5v",
                    "5570": "g8v",
                    "4850": "g5i",
                    "5250": "k0v",
                    "4560": "k4v",
                    "4060": "k7v",
                    "3500": "m2v"}

CK04_MODEL_DICT = { "45000__4.5": "o3v",
                    "41000__4.5": "o5v",
                    "37000__4.0": "o7v",
                    "33000__4.0": "o9v",
                    "30000__4.0": "b0v",
                    "25000__4.0": "b1v",
                    "19000__4.0": "b3v",
                    "15000__4.0": "b5v",
                    "12000__4.0": "b8v",
                    "9500__4.0": "a0v",
                    "9250__4.0": "a1v",
                    "8750__4.0": "a3v",
                    "8250__4.0": "a5v",
                    "7250__4.0": "f0v",
                    "7000__4.0": "f2v",
                    "6500__4.0": "f5v",
                    "6250__4.0": "f8v",
                    "6000__4.5": "g0v",
                    "5750__4.5": "g2v",
                    "5500__4.5": "g8v",
                    "5250__4.5": "k0v",
                    "4750__4.5": "k2v",
                    "4250__4.5": "k5v",
                    "4000__4.5": "k7v",
                    "3750__4.5": "m0v",
                    "3500__4.5": "m2v",
                    "3500__5.0": "m6v",
                    "29000__3.5": "b0iii",
                    "15000__3.5": "b5iii",
                    "5750__3.0": "g0iii",
                    "5250__2.5": "g5iii",
                    "4750__2.0": "k0iii",
                    "4000__1.5": "k5iii",
                    "3750__1.5": "m0iii",
                    "26000__3.0": "b0i",
                    "14000__2.5": "b5i",
                    "9750__2.0": "a0i",
                    "8500__2.0": "a5i",
                    "7750__2.0": "f0i",
                    "7000__1.5": "f5i",
                    "5500__1.0": "g0i",
                    "4750__1.0": "g5i",
                    "4500__1.0": "k0i",
                    "3850__0.0": "k5i",
                    "3650__0.0": "m0i",
                    "3600__0.0": "m2i"
                    }

ACQ_MODE_DICT = { "acq/peak": "acq_peak",
                  "acq": "acq_image"
}

MAMA_DETECTORS = [  ['acs', 'sbc'],
                    ['stis', 'fuvmama'],
                    ['stis', 'nuvmama'],
                    ['cos', 'fuv'],
                    ['cos', 'nuv']]

IGNORE_KEYS = '[ \"scalar.saturation_time\", \"scalar.reference_wavelength\"]'

# unlike peng2jeng, we are not tying ourselves to only translate and produce currently-valid ETC calculations.
# though we should have a flag that will only print valid calculations as determined by successful translation 
# and/or appearance of the value in the relevant config.json

# In this new construction, the second item in the list is either the location in the .jeng dictionary OR 
# the name of a function that will parse that input.
# The same construction as before holds: 
# 1. We load the .peng file, then parse each line to build up a list of what to parse.
# 2. Then we create a default .jeng calculation dictionary
# 3. All the simple translations are just added to the dictionary
# 4. We go through all the '_parse' functions to populate things that need translation or to read from multiple fields
# 5. Then we go through the '_valid' functions to check the original value against the new one.

def get_instdotmode_from_output(str_blob):
    ''' grab the instrument+mode value out of the output - bit of a tiny hack for reporting '''
    assert 'Wrote ' in str_blob
    assert ' test file: ' in str_blob
    for line in str_blob.split('\n'):
        if line.startswith('Wrote ') and ' test file: ' in line:
            # grab idm = inst-dot-mode (e.g. "stis.echelle")
            idm = line.split(' ')[1]
            inst = idm.split('.')[0]
            #assert inst in HST_INSTS + HST_INSTS_NEW_NAMES, 'unexpected inst name: %s' % inst
            return idm
    assert False, 'Expected string not found in the str_blob: %s' % str_blob


def safe_eval(astr):
    '''
    this is a stand-in for pyetc.main.util:safe_eval() because we do NOT want to pull all of
    that in here simply to load a dictionary file...
    '''
    lines = astr.split('\n')
    nocomments = [l for l in lines if not l.strip().startswith('#')]
    buf = ' '.join(nocomments)
    buf = buf.replace('\t',' ') # the result is valid Python, but not yet valid JSON

    # NOTE - since this is not public facing, we are going to use eval here.  This
    # is a temporary tool and will be dropped after the conversion to pandeia.
    return eval(buf)


def _safe_update(insdict, key, value):
    if key in insdict:
        assert insdict[key] == value, f"Inconsistent values for {key}: {insdict[key]}, {value} "
    insdict[key] = value
    return insdict

def pyetc_main_util_read_dict(fname):
    '''
    (stolen directly from pyetc.main.util so as to not require a depenency on pyetc if possible)
    read a python dictionary from a file that was written with write_dict.
    '''
    f=open(fname,'r', encoding="utf-8")
    datastr = f.read()
    f.close()
    # convert DOS file to Unix - otherwise the eval will fail
    datastr = datastr.replace('\r','')
    try :
        datadict = safe_eval(datastr)
    except Exception as e:
        raise ValueError('Exception: %s, from data file: %s' % (e, fname))
    return datadict


def peng2dict(fname):
    " convenience function to load an input file "
    assert os.path.exists(fname), 'Input file unfound: %s' % fname
    # can't just use json.load for pyetc peng files
    return pyetc_main_util_read_dict(fname)

# The correctors, for complex translations

def disperser_valid(adict, calculation):
    assert calculation['configuration']['instrument']['disperser'] in adict['disperser'] or calculation['configuration']['instrument']['aperture'] in adict['disperser'], f"Inconsistent disperser config {adict['disperser']}"
    return calculation

def detector_parse(adict, calculation, pweb_key):
    # We need to settle once and for all if we're making detector independent of aperture. This and obsmode_parse disagree on that point.
    # Basically, we are... except for WFC3 UVIS1/UVIS2, which are so similar they will share modes.

    # See if we need to translate the mode name, otherwise use what's there already.
    mode_name = MODE_NAME_DICT.get(adict['science_mode'], adict['science_mode'])

    if 'uvis' in adict['extraction']['detector']:
        calculation['configuration']['instrument']['mode'] = f"uvis_{mode_name}"
    else:
        calculation['configuration']['instrument']['mode'] = f"{adict['extraction']['detector']}_{mode_name}"


    calculation['configuration']['instrument']['aperture'] = adict['extraction']['detector']
    if 'aperture' in adict:
        if adict['aperture'] != adict['extraction']['detector'] and adict['instrument'] != 'stis':
            # STIS pengs have the filter in the aperture keyword.  Just leave aperture as extraction/detector.
            calculation['configuration']['instrument']['aperture'] = adict['aperture']
    assert adict['detector'] == adict['extraction']['detector']

    # specific for COS spec acq mode
    if 'cos' in adict['instrument'] and 'spec' in adict['science_mode']:
        calculation['configuration']['detector']['gain'] = "1"
        calculation['configuration']['detector']['nsplit'] = 1
    # COS Imaging and ImagingAcq both require the aperture to be set to a mirror
    elif "cos" in adict["instrument"] and ("imaging" in adict['science_mode'] or "targetacquisition" in adict["science_mode"]):
        calculation['configuration']['instrument']['aperture'] = adict['mirror']

    return calculation

def obswave_parse(adict, calculation, pweb_key):
    # Convert the value from Angstroms to microns and store in ['strategy']['reference_wavelength']
    val = float(adict['obswave']) * u.angstrom
    calculation['strategy']['reference_wavelength'] = val.to_value(u.micron)

    return calculation

def obsmode_parse(adict, calculation, pweb_key):

    # Instrument-dependent
    # science mode dependent
    # 
    pweb_dict = locate_pweb(pweb_key)
    obsmode_list = adict['obsmode'].split(',')
    inst = calculation['configuration']['instrument']['instrument']
    if 'wfc3' in inst:
        inst = 'wfc3'

    assert inst in obsmode_list[0], "Invalid instrument"
    result = False
    # some STIS obsmode strings have neither filter nor disperser
    # for item in obsmode_list:
    #     if calculation['configuration']['instrument']['filter'] is not None and calculation['configuration']['instrument']['filter'] in item.lower():
    #         result = True
    #     if calculation['configuration']['instrument']['disperser'] is not None and calculation['configuration']['instrument']['disperser'] in item.lower():
    #         result = True
    # assert result

    if "MJD" in obsmode_list[-1]:
        """
        MJD is not required, but is present in virtually all pyetc tests (added 
        by pyetc/etc_web/etc/perform_calculation.py _handle_mjd(), called by
        process_form_input().
        All further notes will consider this to have been removed.
        """
        calculation['configuration']['instrument']['mjd'] = int(obsmode_list.pop(-1).split("#")[1])
    if 'cos' in calculation['configuration']['instrument']['instrument']:
        if 'spectro' in adict['science_mode']:
            """
            instrument,detector,grating,cenwave,aperture
            Identical processing to stis spectroscopy except that the cenwave is always 
            present, and prepended with ‘c’. There is some hardcoded checking in the 
            instrument_from_form function for invalid combinations of grating and stripe
            """
            calculation['configuration']['instrument']['detector'] = obsmode_list[1].lower()
            calculation['configuration']['instrument']['disperser'] = obsmode_list[2].lower()
            calculation['configuration']['instrument']['cenwave'] = obsmode_list[3].lower()
            calculation['configuration']['detector']['nsplit'] = 1

            if obsmode_list[4].lower() in APERTURE_DICT:
                calculation['configuration']['instrument']['aperture'] = \
                    APERTURE_DICT[obsmode_list[4].lower()][adict['detector']]
                calculation['configuration']['instrument']['slit'] = obsmode_list[4].lower()
            else:
                calculation['configuration']['instrument']['aperture'] = obsmode_list[4].lower()
            calculation['configuration']['instrument']['slit'] = obsmode_list[4].lower()

            if 'spectroscopicacq' in adict['science_mode']:
                calculation['strategy']['method'] = 'taspec'

        elif ('targetacquisition' in adict['science_mode']) or ('imaging' in adict['science_mode']):
            """
            instrument,mirror,detector,aperture
            set by make_cos_imaging_obsmode, all of the heavy lifting is done by 
            _get_cos_imag_parameters, which pulls mirror out of 'Mirror' in the pweb file, 
            and aperture out of ‘cosaperture0’ in the pweb file. detector is present and 
            always ‘nuv’. There are no filters.
            """
            calculation['configuration']['instrument']['detector'] = obsmode_list[2].lower()
            calculation['configuration']['instrument']['aperture'] = obsmode_list[1].lower()
            calculation['configuration']['instrument']['slit'] = obsmode_list[3].lower()
            calculation['configuration']['instrument']['filter'] = None
            calculation['configuration']['detector']['nsplit'] = 1
        else:
            raise NameError(f"Could not identify science_mode {adict['science_mode']}")
    elif 'stis' in calculation['configuration']['instrument']['instrument']:
        if 'spectro' in adict['science_mode']:
            """
            instrument,detector,disperser,[cenwave],aperture
            All work handled by make_stis_spectroscopic_obsmode, which calls 
            _get_stis_spec_parameters. The thing that calls make_stis_spectroscopic_parameters 
            then reruns _get_stis_spec_parameters to get values for itself. But! We then pull 
            the aperture values out of the OBSMODE to put in wo_dict; it’s not modified by 
            make_stis_spectroscopic_obsmode so I don’t know why.
            The only optional component is cenwave, which make_stis_spectroscopic_obsmode 
            modifies by prepending a c or i depending on a hardcoded table; cenwave is either 
            [3] or not present. Slit is always last.
            """
            calculation['configuration']['instrument']['detector'] = obsmode_list[1].lower()
            calculation['configuration']['instrument']['disperser'] = obsmode_list[2].lower()
            calculation['configuration']['instrument']['aperture'] = obsmode_list[1].lower()
            calculation['configuration']['instrument']['slit'] = obsmode_list[3].lower()
            
            if len(obsmode_list) > 4:
                calculation['configuration']['instrument']['cenwave'] = obsmode_list[3].lower()
                calculation['configuration']['instrument']['slit'] = obsmode_list[4].lower()

            if calculation['configuration']['instrument']['detector'] != 'ccd':
                # Pandeia NEEDS bin_spatial/dispersion to be removed, even though pyetc is okay with it as long as it's set to 1.
                del calculation['configuration']['detector']['bin_spatial']
                del calculation['configuration']['detector']['bin_dispersion']

            if calculation['configuration']['instrument']['disperser'].startswith('e'):
                # This is an echelle mode calc, fix the input params.
                calculation['configuration']['instrument']['aperture'] = calculation['configuration']['instrument']['aperture'] + 'echelle'
                calculation['configuration']['instrument']['mode'] = calculation['configuration']['instrument']['mode'].replace('spec', 'echelle')
        elif 'targetacquisition' in adict['science_mode']:
            """
            stis targetacquisition: Basically same as imaging, except there is no "mirror"
            item. And if the ccdmode is peak and the aperture/filter has an x in it, we
            remove any '.' in the aperture string and prepend an s. Also, there is no
            “mirror” item. When acq is peak, the aperture IS an aperture (slit) (or rather
            an aperture/filter combo, given that they seem to have ND filters); when acq
            isn’t peak, it’s a filter.
            """
            calculation['configuration']['instrument']['detector'] = obsmode_list[1].lower()
            calculation['strategy']['method'] = 'taboxphot'
            calculation['configuration']['instrument']['slit'] = adict['aperture'].lower()
            calculation['configuration']['instrument']['filter'] = None
            calculation['strategy']['acq_mode'] = ACQ_MODE_DICT[calculation['strategy']['ccdmode']]
            calculation['strategy']['acq_type'] = 'point'
            calculation['strategy']['aperture_size'] = adict['extraction']['size']
            del calculation['strategy']['ccdmode']
        elif 'imaging' in adict['science_mode']:
            """
            instrument,detector,”mirror”,[aperture]
            The work is done by _get_stis_imag_parameters, which determines detector and 
            aperture (which is apparently actually the filter in idict; it’s the aperture 
            in wodict). The other parameter (possibly the actual aperture? or grating? is 
            “mirror”, hardcoded. _stis_imag_obsmode implies that aperture (aka filter) is 
            optional, but given that even the 50CCD/clear position is specified in PyETC 
            obsmode strings (and you can't use an ACTUAL slit/aperture), I suspect it’s 
            NOT optional (for PyETC use, anyway)
            """
            calculation['configuration']['instrument']['detector'] = obsmode_list[1].lower()
            # STIS pengs have the filter in the aperture and obsmode settings.
            #  Aperture is set correctly in the detector parsing function.
            if len(obsmode_list) > 3:
                calculation['configuration']['instrument']['filter'] = obsmode_list[3].lower()

            if calculation['configuration']['instrument']['detector'] != 'ccd':
                # Pandeia NEEDS bin_spatial to be removed, even though pyetc is okay with it as long as it's set to 1.
                del calculation['configuration']['detector']['bin_spatial']
        else:
            raise NameError(f"Could not identify science_mode {adict['science_mode']}")
    elif 'wfc3' in calculation['configuration']['instrument']['instrument']:
        if 'spectro' in adict['science_mode']:
            """
            wfc3ir
            instrument,ir,disperser
            disperser (called filter in the code) comes from the pweb file as 
            ‘disperser’ (made lowercase). The other two items are hardcoded. The 
            only complexity is that we strip the first four characters from 
            ‘disperser’ in the form to get the disperser. The other oddity is that 
            the “grating” in the peng file is the LAST four characters of the 
            pweb[‘disperser’] entry. A quick check of a pweb file shows that the 
            disperser contains an 8-char string (ir_0G102 or similar) so both bits 
            of code just happen to be doing the same thing.
            wfc3uvis
            instrument,detector,g280
            Done by make_wfc3uvis_spectroscopic_obsmode; it only pulls ‘detector’ 
            from the form_input and hardcodes the g280 grating. (it IS the only 
            one, and not selectable)
            """

            calculation['configuration']['instrument']['detector'] = obsmode_list[1].lower()
            calculation['configuration']['instrument']['disperser'] = obsmode_list[2].lower()
        elif "imaging" in adict['science_mode']:
            """
            wfc3ir
            instrument,ir,filter
            filter comes directly from the pweb file as ‘irfilt0’ (made lowercase). 
            The other two items are hardcoded. It’s that simple.
            wfc3uvis
            instrument,detector,filter
            work is done by make_wfc3uvis_imaging_obsmode, which checks the 
            wfc3_filter_type in the form to figure out which form input to read for 
            the filter. Detector is ‘detector’ from the form input.
            """
            calculation['configuration']['instrument']['detector'] = obsmode_list[1].lower()
            calculation['configuration']['instrument']['filter'] = obsmode_list[2].lower()
        else:
            raise NameError(f"Could not identify science_mode {adict['science_mode']}")
    elif 'acs' in calculation['configuration']['instrument']['instrument']:
        if 'spectro' in adict['science_mode']:
            """
            instrument,detector,[disperser],[filter]
            Complex selection methods, because this is sometimes called with an input 
            (peng) dictionary instead of a pweb dictionary; sometimes the detector and 
            grating are separate items (peng), sometimes they’re combined into a single 
            ‘disperser’ string (pweb) with detector_disperser, and the form has to be 
            parsed to figure out which detector’s filter to pull out (with the 
            additional complication that hrc’s listings are backwards - if the disperser 
            starts with 0, look at hrcfilt1; if the disperser starts with 1, look at 
            hrcfilt0). There can either be both or neither of them if any are “clear”. 
            (though the web interface doesn’t allow the disperser to be clear, so it’s 
            effectively required.) Left unsaid: some of the filters are actually 
            polarizers. (of course, given that neither ETC intends to actually implement 
            polarization, they are effectively filters)
            """
            det = adict['detector']
            if det in APERTURE_DICT:
                calculation['configuration']['instrument']['aperture'] = \
                    APERTURE_DICT[det][det]
                calculation['configuration']['instrument']['detector'] = \
                    APERTURE_DICT[det][det]
            else:
                calculation['configuration']['instrument']['aperture'] = obsmode_list[1].lower()
            calculation['configuration']['instrument']['detector'] = obsmode_list[1].lower()

            # According to ACS Handbook, dispersers start with 'G8' or 'PR'.
            # Polarizers start with 'POL'
            # Everything else will be assumed to be a filter.

            # Loop through the remaining obsmode_list items, and determine if they are filter/polarizer/disperser...
            filter_first = ''
            filter_second = ''
            for i in range(len(obsmode_list) - 2): #[0] and [1] are instrument and detector, so skip the first 2 items.
                index = i + 2
                if obsmode_list[index][:2] in ['g8', 'pr']:
                    filter_first = obsmode_list[index].lower()
                elif str.startswith(obsmode_list[index], 'pol'):
                    filter_second = obsmode_list[index].lower()
                else:
                    filter_first = obsmode_list[index].lower()

            if filter_second != '':
                calculation['configuration']['instrument']['disperser'] = filter_first + ',' + filter_second
            else:
                calculation['configuration']['instrument']['disperser'] = filter_first

        elif 'ramp' in adict['science_mode']:
            """
            instrument,detector,filter#wave
            As with acs imaging, the ETC assumes any use of wfc is using wfc1 as the detector. 
            Filter is read out of the pweb file as ‘(detector)filt1’, and the obswave is read 
            from ‘obswave’ and concatenated. The problem here is that pysynphot has a bug (at 
            least as of 7 years ago) in that it couldn’t support the full wavelength range of 
            the ramp filters, and the solution was to actually run pysynphot on the obsmode 
            string to see if it crashed, and then tell the user the wavelength is out of 
            range. Presumably this script will never run into one of those because those .peng
            files could never be made.
            """
            calculation['configuration']['instrument']['detector'] = obsmode_list[1].lower()
            if obsmode_list[1] == 'wfc1':
                calculation['configuration']['instrument']['detector'] = 'wfc1'
            calculation['configuration']['instrument']['filter'] = obsmode_list[2].split('#')[0].lower()
            calculation['strategy']['reference_wavelength'] = (float(obsmode_list[2].split('#')[1]) * u.angstrom).to_value(u.micron)
        elif 'imaging' in adict['science_mode']:
            """
            instrument,detector,[filter1],[filter2]
            Very complex processing, mostly done in make_acs_imaging_obsmode. wfc is split 
            across two detectors, but ETC assumes any wfc is using the wfc1 detector. hrc 
            and wfc can have two filters (taken from ‘(detector)filt0’ and ‘(detector)filt1’ 
            in the pweb dictionary. Of the input filters, the next element is any non-clear 
            filter(s), though later code throws an error if you have more than one non-clear 
            filter (which already shouldn’t be in the list of filters you created earlier) 
            There does not seem to be a prescribed order of filter and polarizer or filter 
            and clear (though it’s likely that, because all the polarizers are on wheel 2, 
            that they’re all the second item); a separate block of code finds and returns 
            them to the acs instrument_from_form() function. That block of code does make 
            it clear that all polarizers start with “pol” and all clear filters start with 
            “clear”. But again, the code should be weeding out clear filters, so you could 
            have a case where both are clear and the obsmode string ends at “detector”. Using 
            clear in both wheels is allowed, but generates a not-supported warning.
            """
            detector = obsmode_list[1].lower()
            if detector == 'wfc':
                detector = 'wfc1'
            calculation['configuration']['instrument']['detector'] = detector
            calculation['configuration']['instrument']['aperture'] = detector

            # ACS WFC Imaging has an option of a CLEAR filter that's actually not present in the obsmode string at all.
            if len(obsmode_list) > 2:
                calculation['configuration']['instrument']['filter'] = obsmode_list[2]
            if len(obsmode_list) > 3:
                calculation['configuration']['instrument']['filter'] += ',' + obsmode_list[3]

    return calculation


ACQ_MODE_TRANSLATION = {
    'ACQ': 'acq_search',
    'ACQ/PEAKD': 'acq_peakd',
    'ACQ/PEAKXD': 'acq_peakxd'
}

def extraction_parse(adict, calculation, pweb_key):
    pweb_dict = locate_pweb(pweb_key)
    #calculation['strategy']['gyromode'] = adict['extraction']['gyromode']
    calculation['configuration']['instrument']['detector'] = adict['extraction']['detector']
    calculation['strategy']['background_subtraction'] = False

    # Calculate the aperture region for background subtraction.  It should have the same area as
    #  the pyetc source extraction region, but be located away from the source region.
    #inner_edge = adict['extraction']['size'] + 0.5
    #outer_edge = np.sqrt(inner_edge**2.0 + adict['extraction']['size']**2.0)
    #calculation['strategy']['sky_annulus'] = [inner_edge, outer_edge]

    del calculation['strategy']['sky_annulus']

    # Pandeia expects the aperture_size parameter to be in arcsec.
    #  We need to read in the plate scale to convert pixels to arcsec.
    detector = adict['extraction']['detector']
    instrument = DETECTOR_INSTRUMENT_DICT[detector]
    with open(DATA_DIR + '/hst/%s/config.json' % instrument, 'r') as config_file:
        config_data = config_file.read()

    config = json.loads(config_data)

    # To get the plate scale, we need to know the aperture.  To get the aperture, set up an
    #  Instrument object.
    inst_config = {}
    inst_config['instrument'] = {}
    inst_config['instrument']['instrument'] = instrument
    # WFC3 UVIS1 and UVIS2 are so similar, they will share a mode.
    # See if we need to translate the mode name, otherwise use what's there already.
    mode_name = MODE_NAME_DICT.get(adict['science_mode'], adict['science_mode'])

    if 'uvis' in adict['extraction']['detector']:
        inst_config['instrument']['mode'] = f"uvis_{mode_name}"
    elif 'ir' in adict['extraction']['detector']:
        inst_config['instrument']['mode'] = f"ir_{mode_name}"
    elif instrument == 'stis' and 'spec' in adict['science_mode']:
        if adict['grating'].startswith('e'):
            # Echelle mode
            inst_config['instrument']['mode'] = f"{detector}_echelle"
        elif adict['grating'].startswith('g') or adict['grating'] == "prism":
            # 1st order spec mode.  Slit and slitless are in the same modes
            inst_config['instrument']['mode'] = f"{detector}_{mode_name}"
    elif instrument == 'cos' and 'spec' in adict['science_mode'] and detector == 'fuv':
        inst_config['instrument']['mode'] = f"{detector}_spec"
    else:
        inst_config['instrument']['mode'] = f"{detector}_{mode_name}"

    if 'filter' in adict.keys():
        if "clear" in adict["filter"]: # specifically, clear2l filter used by ACS WFC is actually not-appearing-in-the-obsmode-string.
            pass
        else:
            inst_config['instrument']['filter'] = adict['filter'].lower()

    # Creating a STIS FUVMAMA or CCD instrument requires an extra param.
    #  Give it a value here (doesn't matter which for conversion purposes).
    if 'detector' in adict.keys() and 'stis' in instrument:
        inst_config['detector'] = {}
        if 'fuv' in adict['detector']:
            inst_config['detector']['fuv_glow_region'] = 'low'
        elif 'ccd' in adict['detector']:
            inst_config['detector']['dark_level'] = 'low'

    print(inst_config)
    inst = InstrumentFactory(config=inst_config)
    aperture = inst.get_aperture()

    plate_scale_x = config['aperture_config'][aperture]['plate_scale_x'] # arcsec / pixel
    plate_scale_y = config['aperture_config'][aperture]['plate_scale_y'] # arcsec / pixel

    # Use the shape parameter and the science_mode parameter to determine the resulting strategy in the JENG file.
    if adict['extraction']['shape'] == 'rectangle' and 'spectro' not in adict['science_mode']:
        if 'scimag' in adict['science_mode']:
            calculation['strategy']['method'] = 'imagingscanapphot'
            if adict['extraction']['size'][1] == 1.0:
                calculation['strategy']['scan_extraction_length'] = 'pixel'
            else:
                calculation['strategy']['scan_extraction_width'] = 'scanlength'
            calculation['strategy']['scan_extraction_width'] = adict['extraction']['size'][0]
            del calculation['strategy']['aperture_size']
        else:
            raise Exception('ImagingRectPhot not yet implemented in Pandeia')
            calculation['strategy']['method'] = 'imagingrectphot'
            calculation['strategy']['display_string'] = "Imaging Rectangular Photometry"
            calculation['strategy']['aperture_size'] = [adict['extraction']['size'][0] * plate_scale_x, adict['extraction']['size'][1] * plate_scale_y]
    elif adict['extraction']['shape'] == 'rectangle':
        if 'scspectro' in adict['science_mode']:
            calculation['strategy']['method'] = 'specscanapphot'
            if adict['extraction']['size'][0] == 1.0 and adict['extraction']['size'][1] == 1.0:
                calculation['strategy']['scan_extraction_length'] = "pixel"
            else:
                calculation['strategy']['scan_extraction_length'] = "scanlength"
        else:
            if adict['instrument'] == 'cos' and adict['detector'] == 'fuv':
                calculation['strategy']['method'] = 'specsegmentedphot'
            else:
                calculation['strategy']['method'] = 'specapphot' # note that SpecApPhot only accepts the half-height of the box, and relies on other means to get the length of the extraction box
        if 'spectro' in adict['science_mode']:
            if adict['instrument'] == 'wfc3ir' and 'scspec' in adict['science_mode']:
                # WFC3 IR Scan Spec can't have aperture_size in it.

                del calculation['strategy']['aperture_size']
            else:
                calculation['strategy']['aperture_size'] = adict['extraction']['size'][0] * plate_scale_x
        else:
            calculation['strategy']['aperture_size'] = (adict['extraction']['size'][0] * plate_scale_x, adict['extraction']['size'][1] * plate_scale_y)
        calculation['strategy']['display_string'] = "Aperture Spectral Extraction"
    elif adict['extraction']['shape'] == 'square':
        if adict['science_mode'] == 'targetacquisition':
            calculation['strategy']['method'] = 'tarectphot'
            calculation['aperture_size'] = adict['extraction']['size'] * plate_scale_x
            calculation['strategy']['display_string'] = "Imaging Target Acquisition"
            if 'cosAcqMode' in pweb_dict and 'image' == pweb_dict['cosAcqMode']:
                calculation['strategy']['acq_mode'] = 'acq_image'
            del calculation['strategy']['aperture_size']
        else:
            raise Exception('ImagingRectPhot not yet implemented in Pandeia')
            calculation['strategy']['method'] = 'imagingrectphot'
            calculation['strategy']['aperture_size'] = (adict['extraction']['size'] * plate_scale_x,adict['extraction']['size'] * plate_scale_y)
            calculation['strategy']['display_string'] = "Imaging Rectangular Photometry"
    elif adict['extraction']['shape'] == 'circle':
        calculation['strategy']['method'] = 'imagingapphot'
        calculation['strategy']['aperture_size'] = float(adict['extraction']['size'])
        calculation['strategy']['is_aperture_ee'] = False
    elif adict['extraction']['shape'] == 'percent':
        calculation['strategy']['method'] = 'imagingapphot'
        calculation['strategy']['aperture_size'] = float(adict['extraction']['size'])
        calculation['strategy']['is_aperture_ee'] = True
        calculation['strategy']['units'] = 'arcsec'
    if 'stripe' in adict['extraction']:
        calculation['strategy']['stripe'] = adict['extraction']['stripe'].lower()
    if adict['extraction']['stype'] == 'extended':
        # According to Ivo, sdiameter and fdiameter are the same, and pyetc uses compatibility code like this
        try:
            diameter = adict['extraction']['sdiameter']
        except KeyError:
            try:
                diameter = adict['extraction']['fdiameter']
            except KeyError:
                diameter = adict['fdiameter']
        calculation['scene'][0]['shape'] = {"geometry": "flat","major": diameter, "minor": diameter,
            "norm_method": "surf_center", "surf_area_units": "arcsec^2" }

    # COS spec acq mode uses an extra parameter, acq_mode
    if 'acq_mode' in adict['extraction']:
        acq_mode = adict['extraction']['acq_mode']
        calculation['strategy']['acq_mode'] = ACQ_MODE_TRANSLATION[acq_mode]

    return calculation

def simmode_valid(adict, calculation):
    # as per https://innerspace.stsci.edu/display/JEP/2021-11-04+Reverse+Calc+API+Meeting+notes

    calculation['configuration']['detector']['time'] = adict['time']
    calculation['configuration']['detector']['snr'] = adict['SNR']
    calculation['configuration']['detector']['calculate_snr'] = adict['simmode'] == 'SNR'

    return calculation

def locate_pweb(pweb_key):
    """
    Returns the pweb dict given a "pweb key" which is either a peng fname (for which it is assumed to
    have a matching pweb file that we can find), OR it is the pweb dict itself.  In the former
    case we find the file, read it and return the dict.  In the latter case we just return the dict.

    This is generalized to be multipurpose.

    File path searching assumes the peng is in pyetc/test/engine.

    Returns None (without error) if no such file is found.  Callers chack this.
    """
    if isinstance(pweb_key, dict):
        return pweb_key # this is the pweb dict

    pweb_fname = pweb_key.replace('test/engine', 'test/web').replace('.peng', '.pweb')
    pweb_dict = None
    try:
        # load the pweb file
        with open(pweb_fname, 'r') as infile:
            pweb_dict = eval(infile.read())
    except FileNotFoundError:
        pass # pweb_dict = None

    return pweb_dict

def sourceval_parse(adict, calculation, pweb_key):
    sky_expr = adict['source_expr']
    pweb_src_keys = ['spec(', 'z(', 'icat', 'unit(', 'pl(', 'bb(', 'em(', 'rn(']
    try:
        # If this source_expr contains anything that might be easier to parse with a pweb, try and locate the pweb.
        if any(key in sky_expr for key in pweb_src_keys):
            pweb_dict = locate_pweb(pweb_key)
            #if 'spectrum' not in calculation['scene'][0]:
            #    calculation['scene'][0]['spectrum'] = {'extinction': {'bandpass': '', 'law': '', 'unit': '', 'value': 0.0}, 'lines': [], 'name': '', 'redshift': 0.0, 'sed': {'sed_type': {}}, 'normalization': {'norm_wave': {}, 'norm_waveunit': {}}}
            #if 'position' not in calculation['scene'][0]:
            #    calculation['scene'][0]['position'] = {'orientation': 0.0, 'x_offset': 0.0, 'y_offset': 0.0}

            if 'spec(' in sky_expr:
                # A spectrum from CDBS
                # key seems to not be stored in the pweb.  Still need to parse this from the peng I guess.
                #  spec always seems to be the innermost statement in a nest, so this shouldn't be difficult.
                sedstart = sky_expr.find('spec(')
                sedend = sky_expr.find(')', sedstart)
                spec = sky_expr[sedstart:sedend].replace('spec(', '')
                parser = spec.split('$')
                if parser[0] == "crgridbz77":
                    spec = 'grid/bz77/' + parser[1]
                elif parser[0] == "crgrid":
                    spec = 'grid/' + parser[1]
                elif parser[0] == "crcalspec":
                    spec = 'calspec/' + parser[1]
                elif parser[0] == "": # starts with $PYSYN_CDBS
                    spec = "/".join(spec.split("/")[1:])
                calculation['scene'][0]['spectrum']['sed'] = {'sed_type': 'custom', 'key': spec}
            if 'z(' in sky_expr:
                calculation['scene'][0]['spectrum']['redshift'] = pweb_dict['fRedshift']
            if 'icat(' in sky_expr:
                # Comparing *189398.peng to *189398.pweb, it's not clear that, teff, metallicity, and log_g values are stored in the pweb.
                #  Parsing the old way for now.
                # A catalog grid to interpolate
                sedstart = sky_expr.find('icat(')
                sedend = sky_expr.find(')', sedstart)
                spec = sky_expr[sedstart:sedend].replace('icat(', '')
                spec_list = spec.split(',')
                #calculation['scene'][0]['spectrum']['sed'] = {'sed_type': spec_list[0], 'teff': float(spec_list[1]),
                #                                            'metallicity': float(spec_list[2]),
                #                                            'log_g': float(spec_list[3])}
                if spec_list[0] == 'k93models':
                    calculation['scene'][0]['spectrum']['sed']['sed_type'] = spec_list[0]
                    calculation['scene'][0]['spectrum']['sed']['key'] = K93_MODEL_DICT[spec_list[1].replace('.', '')]
                    if 'unit' in calculation['scene'][0]['spectrum']['sed']:
                        del calculation['scene'][0]['spectrum']['sed']['unit']
                    if 'z' in calculation['scene'][0]['spectrum']['sed']:
                        del calculation['scene'][0]['spectrum']['sed']['z']
                elif spec_list[0] == 'ck04models':
                    calculation['scene'][0]['spectrum']['sed']['sed_type'] = spec_list[0]
                    calculation['scene'][0]['spectrum']['sed']['key'] = CK04_MODEL_DICT[f"{spec_list[1].replace('.', '')}__{spec_list[3]}"]
                    if 'unit' in calculation['scene'][0]['spectrum']['sed']:
                        del calculation['scene'][0]['spectrum']['sed']['unit']
                    if 'z' in calculation['scene'][0]['spectrum']['sed']:
                        del calculation['scene'][0]['spectrum']['sed']['z']

                if spec_list[0] in UNSUPPORTED_SED_TYPES:
                    raise Exception('Unsupported SED type')
            if 'unit(' in sky_expr:
                #calculation['scene'][0]['spectrum']['sed'] = {'sed_type': 'flat', 'unit': "flam" if pweb_dict['fIsLambda'] == "'true'" else "fnu"}
                sedstart = sky_expr.find('unit(')
                sedend = sky_expr.find(')', sedstart)
                spec = sky_expr[sedstart:sedend].replace('unit(','')
                speclist = spec.split(',')
                if speclist[0] == '1.':
                    calculation['scene'][0]['spectrum']['sed'] = {'sed_type': 'flat', 'unit': speclist[1]}
            if 'pl(' in sky_expr:
                calculation['scene'][0]['spectrum']['sed'] = {'sed_type': 'powerlaw', 'index': pweb_dict['fIndex'], 'unit': pweb_dict['rn_flux_lambda_units']}
            if 'bb(' in sky_expr:
                calculation['scene'][0]['spectrum']['sed'] = {'sed_type': 'blackbody', 'temp': pweb_dict['fbbtemp']}
            if 'em(' in sky_expr:
                # Multiple lines can be added.  Split by '+' and add a line for each 'em('
                #  Example expr: rn(unit(1.,flam),band(galex,fuv),25.000000,abmag)+em(2000,10,1.5e-13,flam)+em(2020,10,1.5e-13,flam)+em(2040,10,1.5e-13,flam)
                # Info not stored in pweb, parse it out the old way here.
                pieces = sky_expr.split('+')
                for piece in pieces:
                    if 'em(' in piece:
                        param_list = piece.replace('em(', '').replace(')', '').split(',')
                        line = {'center': (float(param_list[0])*u.angstrom).to_value(u.micron), 'emission_or_absorption': 'emission', 'profile': 'gaussian', \
                        'strength': syn.units.convert_flux(float(param_list[0]) * u.angstrom, float(param_list[2])*syn.units.FLAM, syn.units.FNU), 'width': (float(param_list[1])/2)/float(param_list[0]) * 3e8}
                        calculation['scene'][0]['spectrum']['lines'].append(line)
            if 'rn(' in sky_expr:
                if pweb_dict['rn_flux_bandpass_units'] in sky_expr:
                    calculation['scene'][0]['spectrum']['normalization']['norm_flux'] = float(pweb_dict['rn_flux_bandpass'])
                    calculation['scene'][0]['spectrum']['normalization']['norm_fluxunit'] = pweb_dict['rn_flux_bandpass_units']
                else:
                    calculation['scene'][0]['spectrum']['normalization']['norm_flux'] = float(pweb_dict['rn_flux_lambda'])
                    calculation['scene'][0]['spectrum']['normalization']['norm_fluxunit'] = pweb_dict['rn_flux_lambda_units']
                    

                if 'band' in sky_expr:
                    normtype = pweb_dict['fftype_filters'].split('.')[1]
                    if normtype == "observation":
                        calculation['scene'][0]['spectrum']['normalization']['bandpass'] = ",".join(adict["obsmode"].split(",")[:-1])
                    else:
                        calculation['scene'][0]['spectrum']['normalization']['bandpass'] = PWEB_FILTER_DICT[pweb_dict[f'filter.{normtype}']]
                    if 'acs' in normtype or "wfc3" in normtype or "nicmos" in normtype:
                        calculation['scene'][0]['spectrum']['normalization']['type'] = "hst"
                    else:
                        calculation['scene'][0]['spectrum']['normalization']['type'] = "photsys"
                    del calculation['scene'][0]['spectrum']['normalization']['norm_wave']
                    del calculation['scene'][0]['spectrum']['normalization']['norm_waveunit']
                    normstart = sky_expr.find('band(')
                if 'box' in sky_expr:
                    # In pyetc, box refers to the normalization option that let's you specify flux at a given wavelength.
                    # In Pandeia, this is the Normalize at Wavelength option.
                    calculation['scene'][0]['spectrum']['normalization']['norm_flux'] = pweb_dict['rn_flux_lambda']
                    calculation['scene'][0]['spectrum']['normalization']['norm_fluxunit'] = pweb_dict['rn_flux_lambda_units']
                    calculation['scene'][0]['spectrum']['normalization']['norm_wave'] = pweb_dict['rn_lambda']
                    calculation['scene'][0]['spectrum']['normalization']['norm_waveunit'] = PWEB_NORMUNIT_DICT[pweb_dict['rn_lambda_units']]
                    calculation['scene'][0]['spectrum']['normalization']['type'] = "at_lambda"
                    normstart = sky_expr.find('box(')
                if 'ebmvx' in sky_expr:
                    extinction_band = PWEB_FILTER_DICT[pweb_dict['filter.ubvri']].split(',')[-1]
                    if extinction_band not in ['v', 'j', 'h', 'k']:
                        extinction_band = 'v'
                    calculation['scene'][0]['spectrum']['extinction']['bandpass'] = extinction_band#'v'
                    calculation['scene'][0]['spectrum']['extinction']['value'] = 'mag'
                    calculation['scene'][0]['spectrum']['extinction']['law'] = EXTINCTION_DICT[pweb_dict['febmvtype']]

                    calculation['scene'][0]['spectrum']['extinction']['value'] = pweb_dict['febv']

                    calculation['calculation']['settings'] = {}
                    if pweb_dict['fextinctiontype'] == 'after':
                        # Pandeia only supports applying extinction BEFORE renormalization.  This can drastically affect results, so let's bail on converting this one.
                        raise Exception("Can't apply extinction after renormalization") 

    except (FileNotFoundError, TypeError):
        calculation = old_sourceval_parse(adict, calculation) # the old parsing is inferior but does not require a PWEB file

    return calculation

def old_sourceval_parse(adict, calculation):
    sky_expr = adict['source_expr']

    sky_expr = sky_expr.split('+')
    for item in sky_expr:
        # basically, we're going to remove elements from this until we end up at just a renormalize statement
        if '*' in item:
            scaleindex = item.find('*')
            scalefactorstart = item.rfind('(', 0, scaleindex)
            scalefactor = item[scalefactorstart + 1:scaleindex].strip()

            # Figure out what function this scaling is being applied to.
            funcend = item.find('(', scaleindex)
            funcname = item[scaleindex+1:funcend].strip()
            calculation['scene'][0]['spectrum'][SOURCEVAL_FUNC_DICT[funcname]]['scale'] = scalefactor
            item = item.replace(scalefactor, '').replace('*', '')
        if 'spec(' in item:
            # A spectrum from CDBS
            sedstart = item.find('spec(')
            sedend = item.find(')', sedstart)
            spec_orig = item[sedstart:sedend+1]
            spec = item[sedstart:sedend].replace('spec(','')
            parser = spec.split('$')
            if parser[0] == "crgridbz77":
                spec = 'grid/bz77/' + parser[1]
            elif parser[0] == "crgrid":
                spec = 'grid/' + parser[1]
            elif parser[0] == "crcalspec":
                spec = 'calspec/' + parser[1]
            elif parser[0] == "": # starts with $PYSYN_CDBS
                spec = "/".join(spec.split("/")[1:])
            calculation['scene'][0]['spectrum']['sed'] = {'sed_type': 'cdbs', 'key': spec}
            item = item.replace(spec_orig,'')
        if 'z(' in item:
            zstart = item.find('z(')
            zend = item.find(')', zstart)
            z_orig = item[zstart:zend+1]
            z = item[zstart:zend].replace('z(', '').replace(',', '')
            calculation['scene'][0]['spectrum']['redshift'] = float(z)
            item = item.replace(z_orig, '')
        if 'icat(' in item:
            # A catalog grid to interpolate
            sedstart = item.find('icat(')
            sedend = item.find(')', sedstart)
            spec_orig = item[sedstart:sedend+1]
            spec = item[sedstart:sedend].replace('icat(','')
            speclist = spec.split(',')
            #calculation['scene'][0]['spectrum']['sed'] = {'sed_type': speclist[0], 'teff': float(speclist[1]), 'metallicity': float(speclist[2]), 'log_g': float(speclist[3])}
            if speclist[0] == 'k93models':
                    calculation['scene'][0]['spectrum']['sed']['sed_type'] = speclist[0]
                    calculation['scene'][0]['spectrum']['sed']['key'] = K93_MODEL_DICT[speclist[1]]
                    if 'unit' in calculation['scene'][0]['spectrum']['sed']:
                        del calculation['scene'][0]['spectrum']['sed']['unit']
                    if 'z' in calculation['scene'][0]['spectrum']['sed']:
                        del calculation['scene'][0]['spectrum']['sed']['z']
            item = item.replace(spec_orig,'')
        if 'unit(' in item:
            # Flat spectrum
            sedstart = item.find('unit(')
            sedend = item.find(')', sedstart)
            spec_orig = item[sedstart:sedend+1]
            spec = item[sedstart:sedend].replace('unit(','')
            speclist = spec.split(',')
            if speclist[0] == '1.':
                calculation['scene'][0]['spectrum']['sed'] = {'sed_type': 'flat', 'unit': speclist[1]}
            item = item.replace(spec_orig,'')
        if 'pl(' in item:
            # Power law spectrum
            sedstart = item.find('pl(')
            sedend = item.find(')', sedstart)
            spec_orig = item[sedstart:sedend+1]
            spec = item[sedstart:sedend].replace('pl(','')
            speclist = spec.split(',')
            if speclist[0] == '1.':
                calculation['scene'][0]['spectrum']['sed'] = {'sed_type': 'powerlaw', 'index': -1*float(speclist[1]), 'unit': speclist[2]}
            item = item.replace(spec_orig,'')
        if 'bb(' in item:
            # blackbody spectrum
            sedstart = item.find('bb(')
            sedend = item.find(')', sedstart)
            spec_orig = item[sedstart:sedend+1]
            spec = item[sedstart:sedend].replace('bb(','')
            speclist = spec.split(',')
            calculation['scene'][0]['spectrum']['sed'] = {'sed_type': 'blackbody', 'temp': float(speclist[0])}
            item = item.replace(spec_orig,'')
        if 'em(' in item:
            # emission line
            sedstart = item.find('em(')
            sedend = item.find(')', sedstart)
            spec_orig = item[sedstart:sedend+1]
            spec = item[sedstart:sedend].replace('em(','')
            speclist = spec.split(',')
            line = {'center': (float(speclist[0])*u.angstrom).to_value(u.micron), 'emission_or_absorption': 'emission', 'profile': 'gaussian', \
                'strength': syn.units.convert_flux(float(speclist[0]) * u.angstrom, float(speclist[2])*syn.units.FLAM, syn.units.FNU), 'width': (float(speclist[1])/2)/float(speclist[0]) * 3e8}
            calculation['scene'][0]['spectrum']['lines'].append(line)
            item = item.replace(spec_orig,'')
        if 'rn(' in item:
            # normalize
            if 'band(' in item:
                # normalize to bandpass
                normstart = item.find('band(')
                normend = item.find(')', normstart)
                band_orig = item[normstart:normend+1]
                band = item[normstart:normend].replace('band(','')
                calculation['scene'][0]['spectrum']['normalization']['bandpass'] = band
                calculation['scene'][0]['spectrum']['normalization']['type'] = "photsys"
                item = item.replace(band_orig,'')
                del calculation['scene'][0]['spectrum']['normalization']['norm_wave']
                del calculation['scene'][0]['spectrum']['normalization']['norm_waveunit']
            if 'box(' in item:
                # normalize to wavelength
                normstart = item.find('box(')
                normend = item.find(')', normstart)
                band_orig = item[normstart:normend+1]
                band = item[normstart:normend].replace('box(','')
                bandlist = band.split(',')
                calculation['scene'][0]['spectrum']['normalization']['norm_wave'] = float(bandlist[0])
                calculation['scene'][0]['spectrum']['normalization']['norm_waveunit'] = 'angstroms'
                item = item.replace(band_orig, '')
            if 'ebmvx(' in item:
                print('^^^^^^^^^^^^^^^^^ WHY??????? ^^^^^^^^^^^^^^^^^^^^^')
                # reddening
                redstart = item.find('ebmvx(')
                redend = item.find(')', redstart)
                #rnstart = item.find('rn(')
                #if redstart > rnstart:
                #        # Pandeia only supports applying extinction BEFORE renormalization.  This can drastically affect results, so let's bail on converting this one.
                #        raise Exception("Can't apply extinction after renormalization") 

                red_orig = item[redstart:redend+1]
                red = item[redstart:redend].replace('ebmvx(','')
                redlist = red.split(',')
                val = float(redlist[0])
                calculation['scene'][0]['spectrum']['extinction']['bandpass'] = 'v'
                calculation['scene'][0]['spectrum']['extinction']['unit'] = 'mag'
                # TODO: We have no translations for some of these reddening laws
                if redlist[1] == "mwavg":
                    calculation['scene'][0]['spectrum']['extinction']['law'] = "mwrv31"
                    calculation['scene'][0]['spectrum']['extinction']['value'] = val*3.1
                elif redlist[1] == "mwdense":
                    calculation['scene'][0]['spectrum']['extinction']['law'] = "mwrv55" # this law is Rv = 5.5, but close enough?
                    calculation['scene'][0]['spectrum']['extinction']['value'] = val*5.0
                elif redlist[1] == "mwrv21":
                    calculation['scene'][0]['spectrum']['extinction']['law'] = ""
                    calculation['scene'][0]['spectrum']['extinction']['value'] = val*2.1
                elif redlist[1] == "mwrv4":
                    calculation['scene'][0]['spectrum']['extinction']['law'] = "mwrv40"
                    calculation['scene'][0]['spectrum']['extinction']['value'] = val*4.0
                elif redlist[1] == "lmcavg":
                    calculation['scene'][0]['spectrum']['extinction']['law'] = "lmcavg"
                    calculation['scene'][0]['spectrum']['extinction']['value'] = val*3.41
                elif redlist[1] == "lmc30dor":
                    calculation['scene'][0]['spectrum']['extinction']['law'] = ""
                    calculation['scene'][0]['spectrum']['extinction']['value'] = val*2.76
                elif redlist[1] == "smcbar":
                    calculation['scene'][0]['spectrum']['extinction']['law'] = "smcbar"
                    calculation['scene'][0]['spectrum']['extinction']['value'] = val*2.74
                elif redlist[1] == "xgalsb":
                    calculation['scene'][0]['spectrum']['extinction']['law'] = ""
                    calculation['scene'][0]['spectrum']['extinction']['value'] = val*4.05 # pulled out of the Calzetti et al. (2000. ApJ, 533, 682) paper

                calculation['calculation']['settings'] = {}
                if redend > normstart:
                    # I manually tested this, and it seems to work.  Changing the order of the reddening and band resulted in
                    #   different extinction_first values.
                    # if the reddening ended after the normalization band definition was started, we applied reddening later
                    # the normalization will have been removed when the reddening is defined, but this won't fail because the
                    # normalization entry isn't fully removed, and thus reddening couldn't start before it did.
                    calculation['calculation']['settings']['extinction_first'] = False
                else:
                    calculation['calculation']['settings']['extinction_first'] = True
                item = item.replace(red_orig,'')
            norm = item.replace(')','').split(',')
            flux = float(norm[2])
            flux_units = norm[3]
            # TODO: Revisit this unit conversion.  Pandeia supports vegamag.
            if (flux_units == 'vegamag') and (calculation['scene'][0]['spectrum']['normalization']['type'] == "at_lambda"):
                #Need to convert vegamag to mjy to normalize at_lambda
                sp = pysynphot.ArraySpectrum(np.array([calculation['scene'][0]['spectrum']['normalization']['norm_wave']]), np.array([flux]))
                sp.convert('jy')
                flux = sp.flux[0] / 1e6
                flux_units = 'mjy'
            calculation['scene'][0]['spectrum']['normalization']['norm_flux'] = flux
            calculation['scene'][0]['spectrum']['normalization']['norm_fluxunit'] = flux_units

    return calculation

def dummy_parse(adict, calculation, pweb_key):
    # dummy for now
    return calculation

def lineval_parse(adict, calculation, pweb_key):
    calculation['scene'][0]['spectrum']['lines'] = []
    if adict['target_lines'] is not None:
        for line in adict['target_lines']:
            #fluxunits = FLUX_UNIT_DICT[line['fluxunits']]
            waveunits = WAVE_UNIT_DICT[line['waveunits']]

            # This is derived from the formula for the Doppler shift.  The factor of 1000 converts micrometers to angstroms to make the ratio (width / center) unitless.
            width = ((line['fwhm'] / (line['center'] * 10000.0)) * const.c.to('km/s')).value
            #strength = syn.units.convert_flux(line['center'] * waveunits, line['flux']*fluxunits, syn.units.FNU)
            strength = line['flux']
            center = (line['center'] * waveunits).to_value(u.micron)

            newline = {'center': center, 'emission_or_absorption': 'emission', 'profile': 'gaussian', 'strength': strength, 'width': width}
            calculation['scene'][0]['spectrum']['lines'].append(newline)
    return calculation

def background_parse(adict, calculation, pweb_key):
    sky_expr = adict['sky_expr']

    # Use pysynphot to parse pyetc sky expression.
    sky_expr_parsed = SP.parse_spec(sky_expr)
    sky_expr_parsed.convert('jy')
    sky_expr_parsed.convert('micron')

    # Now assemble the spectrum (flam/arcsec and angstroms) in a format pandeia understands (MJy/sr and microns)
    bgflux = sky_expr_parsed.flux * 1e-6 / (u.arcsec * u.arcsec).to(u.sr)
    #bgflux = sky_expr_parsed.flux / (u.arcsec * u.arcsec).to(u.sr)
    wave = sky_expr_parsed.wave

    calculation['background'] = [wave, bgflux]

    return calculation

def gain_parse(adict, calculation, pweb_key):
    gain = str(adict['gain'])
    if [adict['instrument'], adict['detector']] in MAMA_DETECTORS:
        # ACS SBC (MAMA) doesn't support gain or nreads.  Set to 1.
        gain = "1"

    calculation['configuration']['detector']['gain'] = gain

    return calculation

def nreads_parse(adict, calculation, pweb_key):
    nreads = adict['nreads']
    if [adict['instrument'], adict['detector']] in MAMA_DETECTORS:
        # ACS SBC (MAMA) doesn't support gain or nreads.  Set to 1.
        nreads = 1

    calculation['configuration']['detector']['nsplit'] = nreads

    return calculation

parsefuncs = [dummy_parse, detector_parse, extraction_parse, obsmode_parse, obswave_parse, sourceval_parse, lineval_parse, background_parse, gain_parse, nreads_parse]
validfuncs = [disperser_valid, simmode_valid]

H2J_INPUT_CHECKS = {
#    PENG-name,            JENG-name,                                           data-type,         can-force?,  ...
    'SNR':                 [('configuration','detector','snr'),                 'TYPE:NUM',        False, ],
    'airglow_str':         [dummy_parse,                                        'TYPE:STR',        True ],
    'aperture':            [('configuration','instrument','aperture'),          'TYPE:STR',        False ],
    'binx':                [('configuration','detector','bin_spatial'),         'TYPE:INT',        False ],
    'biny':                [('configuration','detector','bin_dispersion'),      'TYPE:INT',        False ],
    'ccddarklevel':        [('configuration','detector','dark_level'),          'TYPE:STR',        False ],
    'ccdmode':             [('strategy','ccdmode'),                             'TYPE:STR',        False ],
    'central_wavelength':  [dummy_parse,                                        'TYPE:INT',        True ],
    'crsplit':             [('configuration','detector','nsplit'),              'TYPE:INT',        False ], # see also nreads
    'detector':            [detector_parse,                                     HST_DETECTORS,     False ],
    'disperser':           [disperser_valid,                                    'TYPE:STR',        False ],
    'earthshine_str':      [dummy_parse,                                        'TYPE:STR',        True ],
    'etcid':               [dummy_parse,                                        'TYPE:STR',        False ], # will never be used here
    'extraction':          [extraction_parse,                                   'TYPE:DICT',       False ],
    'fdiameter':           [dummy_parse,                                        'TYPE:NUM',        False ],
    'filter':              [('configuration', 'instrument','filter'),           'TYPE:STR',        False ],
    'fuvglowregion':       [('configuration','detector','fuv_glow_region'),     'TYPE:STR',        True ],
    'gain':                [gain_parse,                                         'TYPE:NUM',        False ],
    'grating':             [('configuration','instrument','disperser'),         'TYPE:STR',        False ],
    'helium_str':          [dummy_parse,                                        'TYPE:STR',        True ],
    'instrument':          [('configuration','instrument','instrument'),        HST_INSTS,         False ],
    'nreads':              [nreads_parse,                                       'TYPE:INT',        False ], # see also crsplit
    'obsmode':             [obsmode_parse,                                      'TYPE:STR',        False ],
    'mirror':              [dummy_parse,                                        HST_MIRRORS,       False ],
    'obswave':             [obswave_parse,                                      'TYPE:NUM',        False],
    'post_flash_electrons':[('configuration','detector','postflash'),           'TYPE:INT',        False ],
    'science_mode':        [('configuration','instrument','mode'),              HST_MODES,         False ],
    'sclength':            [('configuration','detector','scan_length'),         'TYPE:NUM',        False ],
    'scrate':              [('configuration','detector','scan_rate'),           'TYPE:NUM',        False ],
    'simmode':             [simmode_valid,                                      'TYPE:STR',        True ], # might force these for a while to pick up more stis tests
    'sky_expr':            [background_parse,                                   'TYPE:STR',        True ],
    'source_expr':         [sourceval_parse,                                    'TYPE:STR',        False ],
    'target_lines':        [lineval_parse,                                      'TYPE:LIST',       True ],
    'time':                [('configuration','detector','time'),                'TYPE:NUM',        False ],
    'zodi_str':            [dummy_parse,                                        'TYPE:STR',        True ],
}

def pyetc_in_dict_to_pandeia_in_dict(adict, input_fname, pweb_dict=None):
    # init
    new_inputs = {}
    msgs = []

    # first pass - go through and warn about unused args/values
    for key in adict:
        # basics
        assert key in H2J_INPUT_CHECKS, 'UNSUPPORTED (as yet!) pyetc input arg: %s' % key
        arg_info = H2J_INPUT_CHECKS[key]
        value = adict[key]
        newname = arg_info[0] if arg_info[0] else key
        valid_vals = arg_info[1]
        assert not valid_vals is None, 'Error in data defined in H2J_INPUT_CHECKS for: %s' % key
        if valid_vals == 'TYPE:DICT':
            assert isinstance(value, dict) or value is None, 'Expected type dict for value of: %s' % key
            # no more checks here yet
            new_inputs[newname] = value
        elif valid_vals == 'TYPE:LIST':
            assert isinstance(value, list) or value is None, 'Expected type list for value of: %s' % key
            # no more checks here yet
            new_inputs[newname] = value
        elif valid_vals == 'TYPE:STR':
            assert isinstance(value, str) or value is None, 'Expected type str for value of: %s' % key
            # no more checks here yet
            if value is not None:
                new_inputs[newname] = value.replace('\n','')
            else:
                new_inputs[newname] = value
        elif valid_vals == 'TYPE:INT':
            assert type(value) == int, 'Expected type int for value of: %s' % key
            # no more checks here yet
            new_inputs[newname] = value
        elif valid_vals == 'TYPE:NUM':
            assert type(value) in (int, float), 'Expected type int or float for value of: %s' % key
            # no more checks here yet
            new_inputs[newname] = value
        else:
            assert type(valid_vals) == list and len(valid_vals) > 0, 'Error in data defined in H2J_INPUT_CHECKS for: %s' % key
            assert value in valid_vals, f'Error in defined data: {newname} {value}'
            new_inputs[newname] = value

    # instrument and mode, filters and dispersers
    calculation = build_default_calc("jwst", "nircam", "sw_imaging")
    calculation = data_util.clean(calculation)
    calculation['background'] = {}
    # some calculations have no filter or disperser
    calculation['configuration']['instrument']['filter'] = None
    calculation['configuration']['instrument']['disperser'] = None
    calculation['configuration']['instrument']['aperture'] = None
    calculation['configuration']['detector'] = {} # Detector is full of JWST-only keys we'll replace.
    del calculation['background_level']
    for item in new_inputs:
        if isinstance(item, tuple):
            if len(item) == 2:
                calculation[item[0]].update({item[1]: new_inputs[item]})
            elif len(item) == 3:
                calculation[item[0]][item[1]].update({item[2]: new_inputs[item]})
        elif isinstance(item, str):
            calculation.update({item: new_inputs[item]})
    for item in new_inputs:
        if item in parsefuncs:
            if pweb_dict:
                calculation = item(adict, calculation, pweb_dict)
            else:
                calculation = item(adict, calculation, input_fname)
    for item in new_inputs:
        if item in validfuncs:
            calculation = item(adict, calculation)

    # merge both pyetc wfc3 instruments to wfc3
    if "wfc3" in calculation['configuration']['instrument']['instrument']:
        calculation['configuration']['instrument']['instrument'] = "wfc3"

    return calculation

def write_sorted_json(data, fname):
    with open(fname, 'w') as f:
        json.dump(data, f, indent=4, separators=(',', ': '), sort_keys=True, cls=NumPyArangeEncoder)
        f.write('\n')

    # Rewrite the ignore_keys.dat file.
    dirname = os.path.dirname(fname)
    local_ignore_keys = IGNORE_KEYS
    if 'cos' in dirname and 'imag' in dirname:
        # Also add 'scalar.disperser' to the list.
        local_ignore_keys = local_ignore_keys.replace(']', ', \"scalar.disperser\"]')
    
    with open(dirname + '/ignore_keys.dat', 'w') as f:
        f.write(local_ignore_keys)

def convert_dict(peng_dict, input_fname):
    pandeia_input_dict = pyetc_in_dict_to_pandeia_in_dict(peng_dict, input_fname)
    return pandeia_input_dict


def convert_file(input_fname, top_out_dir=None, limit_outputs=None):
    """ Convert a single PENG file to a JENG file """
    # output
    # names and paths
    path_part, fname_part = os.path.split(input_fname)
    assert path_part.startswith('/'), 'unexpected relativity in input fname: %s' % input_fname # not absolute
    basename = os.path.splitext(fname_part)[0]
    outstr = '\n'+basename+' ...\n'
    outstr += '-'*80; outstr += '\n'
    out_fname = basename+'.jeng' # for dumping it locally
    # top dir specified for the ouput?
    if top_out_dir:
        unused, path_part_btm = path_part.split('/pyetc/test/')
        path_part = top_out_dir+os.sep+path_part_btm
    if not os.path.exists(path_part):
        os.makedirs(path_part)

    # output name
    out_fname = path_part+os.sep+basename+'.jeng' # for dumping it in place
    # load it to a dict
    pyetc_input_dict = peng2dict(input_fname)
    # convert it
    pandeia_input_dict = convert_dict(pyetc_input_dict, input_fname)


    instdotmode = '%s.%s' % (pandeia_input_dict['configuration']['instrument']['instrument'], pandeia_input_dict['configuration']['instrument']['mode'])

    # see if they asked us to limit the number of test files we create
    if limit_outputs:
        assert limit_outputs.get(instdotmode, 0) < 500, 'Have enough %s tests now' % instdotmode

    # write it out
    if os.path.exists(out_fname):
        os.remove(out_fname)
    write_sorted_json(pandeia_input_dict, out_fname)
    outstr += 'Wrote %s test file: %s\n' % (instdotmode, out_fname)
    return outstr


def run_some_samples(pyetc_test_dir, top_out_dir, single_file=None):
    """
    Select a sampling of all the modes we can run.
    Either send in the first 2 args, or just use the single_file arg.
    This could be smoother, I know.
    """

    # paths under pyetc test
    # places = [
    #     'engine/test_kits/geom-detector', # put these in first so they get grabbed
    #     'engine/bit/stis/imag', # about 250 tests
    #     'engine/bit/wfc3ir/imag/ir',
    #     'engine/sample', # has some COS (not done yet) but some STIS that might work
    #     'engine/spider/acs/imag/sbc-stars-1',
    #     'engine/spider/common/wfc3ir/imag', # this one line adds about 1300 tests
    #     'engine/spider/stis/spec/stellar-pt',
    #     'engine/spider/stis/spec/stellar-pt-bin',
    #     'engine/spider/wfc3ir/imag', # this includes both pt and ext srcs
    #     'engine/spider/wfc3uvis/imag', # this includes both pt and ext srcs
    #     'engine/spider/wfc3ir/scimag',
    #     'engine/test_kits/geom-sim',
    #     'engine/test_kits/limits',
    # ]
    places = [
        'engine/test_kits/', # put these in first so they get grabbed
        'engine/bit/',
        'engine/sample', # has some COS (not done yet) but some STIS that might work
        'engine/spider/'
    ]
    flist = []
    skips = {}

    if single_file:
        flist = [single_file]
        top_out_dir = None # will drop new jeng file right next to old peng file
    else:
        for place in places:
            sublist = utils.rglob(pyetc_test_dir+'/'+place, '*.peng')
            assert len(sublist) > 0, 'no pengs found under %s ??' % place
            flist += sublist

    done = {}

    if single_file:
        convert_file(single_file)
    else:
        # Only use multiprocessing in batch mode to make debugging by converting single files easier.
        with Pool(7) as p:
            for output in p.imap_unordered(convert_file_helper, zip(flist, [top_out_dir]*len(flist))):
                if output[0] == "Success":
                    instdotmode = output[1]
                    fname = output[2]
                    if instdotmode in done:
                        done[instdotmode] += 1
                    else:
                        done[instdotmode] = 1
                    print(f"{fname} Success")
                elif output[0] == "Failure":
                    if len(flist) == 1:
                        raise output[3]
                    reason = output[1]
                    fname = output[2]
                    if reason in skips:
                        skips[reason].append(fname)
                    else:
                        skips[reason] = [fname]
                    print(f"{fname} Failure ({reason}")

    total_skipped = 0
    if skips:
        print("\n\nSKIPS:\n")
    for reason in skips:
        ncases = len(skips[reason])
        print("Skipped %d cases of: %s,  e.g.: %s" % (ncases, reason, skips[reason][0]))
        total_skipped += ncases

    print('\nConverted %d and skipped %d tests. Of the converted: %s' % (len(flist)-total_skipped, total_skipped, done))

def convert_file_helper(convertinput):
    fname, top_out_dir = convertinput
    try:
        outstr = convert_file(fname, top_out_dir)#, limit_outputs=done)
        instdotmode = get_instdotmode_from_output(outstr)
        status = "Success"
        output = (status, instdotmode, fname)
    except Exception as e:
        reason = str(e) if type(e) == AssertionError else repr(e)
        status = "Failure"
        output = (status, reason, fname, e)
    return output


#
# main routine
#
if __name__=='__main__': # in case something else imports this file
    # If this is run with no args, go fetch a same set of tests
    if len(sys.argv) == 1:
        run_some_samples(os.path.expanduser(f'{DATA_DIR}/../pyetc/test'),
                         os.path.expanduser(f'{DATA_DIR}/../pandeia_test/tests/engine_converted/hst'))

    # If this is given a peng file name as an arg on the command line, translate it
    else:
        for fname in sys.argv[1:]:
            run_some_samples(None, None, single_file=fname)

    sys.exit(0)



# Licensed under a 3-clause BSD style license - see LICENSE.rst

import os
from functools import reduce
import numpy as np
from numpy.lib.function_base import iterable
import scipy.sparse as sparse
import operator

import astropy.io.fits as fits
from astropy.convolution import Gaussian1DKernel, convolve

from .detector_factory import DetectorFactory
from . import config as cf
from .psf_library import PSFLibrary
from .io_utils import read_json, ref_data_interp, ref_data_column
from .utils import merge_data, spectrum_resample
from .pandeia_warnings import instrument_warning_messages as warning_messages
from .telescope import TelescopeConfig
from .custom_exceptions import EngineInputError, DataError, UnsupportedError, InternalError, DataConfigurationError
from .constants import DEFAULT_DISPERSION

default_refdata_directory = cf.default_refdata_directory

def notin(i,l):
    return i not in l


def check_warnings(input_dict, warnings, warnings_list, value_lookup, sub_word=None, sub_str=''):
    """
    Input/Output warnings checker.

    The basic way this works is: the instrument config.json file has a warnings section. In it are a series of
    parameters that are interpreted as conditional statements, covering both which instrument setup this
    applies to AND the actual level/value we are checking for. We start with need_warning = True and evaluate
    each condition statements. If any one of them is False, we set need_warning = False and no warning is
    emitted. Warnings are only evaluated if every single one of the statements is True. Thus, we can filter by
    mode, disperser, or filter AND a numerical value.

    Values tested by the conditional are defined by the value_lookup dict as "shortname": [nested dict
    location] and can be inserted into warning strings as $shortname. The values are only evaluated when the
    warning conditions are parsed, so the dict can have parameters that are not defined for all instrument
    setups without fear of crashing. 

    There are two reserved words that must not show up as keys:
    * $threshold: The threshold value that triggers the warning.
    * $subthreshold: The fraction of the threshold value that triggers the subthreshold warning (converted to
      a percentage for output)
    Using these in warning messages should help enforce consistency: we don't have to update the message
    every time the number is changed, nor do we have to worry about copy&paste errors.

    :param warnings_list: list of warnings to check for and add to the warnings list
    :param warnings: warnings dict that is updated by the function with new warnings
    :param input_dict: input as passed into the function
    :param value_lookup: dict with lookup lists for each thing we want to check against the warning parameters
    :param sub_word: str to prefix warning with so that subreports can independently report values
    :param sub_str: string prefix for warning message
    :return:
    """

    # Run through each warning in the warnings_list to see if we need to return it.
    for warning in warnings_list:

        # Run through all the parameters for this warning to see if all are true
        need_warning = True
        thewarning = warning['warning_text']
        thename = warning["name"]
        threshold = None
        for key, expected_value in list(warning['parameters'].items()):

            # Get the value from the json dict. This does a lookup
            # based on the value_lookup above to see where in the input json dict we
            # have to look for the value that was input.
            # https://stackoverflow.com/questions/13687924/setting-a-value-in-a-nested-python-dictionary-given-a-list-of-indices-and-value

            operator_text = '=='
            if '__' in key:
                key, operator_text = key.split('__', 1)

            # https://stackoverflow.com/questions/9320335/accessing-python-dict-with-multiple-key-lookup-string
            input_value = reduce(dict.get, value_lookup[key], input_dict)

            # Setup the comparison operator to use below.
            cmp = '=='
            if operator_text == '==':
                cmp = operator.eq
            elif operator_text == 'ge':
                cmp = operator.ge
            elif operator_text == 'le':
                cmp = operator.le
            elif operator_text == 'gt':
                cmp = operator.gt
            elif operator_text == 'lt':
                cmp = operator.lt
            elif operator_text == 'notin':
                cmp = lambda a,b: a not in b
            else:
                raise InternalError(value="Unknown operator on key {}__{}".format(key, operator_text))

            # Variable-driven threshold value, like fullwell or min_snr_threshold
            if isinstance(expected_value, str) and ("$" in expected_value):
                expected_value = expected_value.replace("$", "")
                expected_value = reduce(dict.get, value_lookup[expected_value], input_dict)

            if operator_text != "==":
                # if this value was subject to a quantitative comparison, save it as the threshold value
                threshold = expected_value

            # For some reason there started to be collisions on comparison of str vs unicode so I am
            # going to convert things to string that might be string or unicode.  That way the comparison
            # will work properly.  In Python3 this will likely be able to be removed.
            if isinstance(input_value, str):
                input_value = str(input_value)
                if isinstance(expected_value, str):
                    expected_value = str(expected_value)
                elif isinstance(expected_value, list):
                    expected_value = [str(x) for x in expected_value]

            # Compare the input_value with the expected value
            if (  # do operator comparison for value
                ### TODO
                ### this line of code prevents crashing when input_value is None
                ### This was put in place to prevent errors with scan_rate.
                ### This should be removed when scan_rate gets implemented.
                ###
                (input_value is not None) and
                (isinstance(expected_value, (str, int, float)) and cmp(input_value, expected_value)) or
                # do set inclusion or exclusion for list
                    (isinstance(expected_value, list) and
                         ((operator_text == '==' and input_value in expected_value) or
                         (operator_text == 'notin' and not input_value in expected_value))
                    )
                ):
                pass  # we do need a warning, based on this test
            elif "subthreshold" in warning and (isinstance(expected_value, (int, float)) and
                                                input_value is not None and
                                                cmp(input_value, expected_value * warning['subthreshold'])):
                ### And now, subthresholds. They are always percentages of a float/int.
                # It makes no sense to have an == or notin operator for a subthreshold.

                # If true, we do need the warning, but the subthreshold warning.
                # Such warnings have the subthreshold percentage in the message; it's in a 
                # Python list formatting syntax for inclusion here.
                thewarning = warning['subthreshold_text']
            else:
                need_warning = False

            # Replace $mode, $instrument, $filter etc in the output warning with the actual value.
            # special syntax if we don't want it to be uppercase.
            if isinstance(input_value, float):
                input_value = f"{input_value:.1f}"
            else:
                input_value = str(input_value)
            if '$~'+key in thewarning:
                thewarning = thewarning.replace('$~' + key, input_value)
            if '$'+key in thewarning:
                thewarning = thewarning.replace('$' + key, input_value.upper())
            if threshold is not None:
                thewarning = thewarning.replace("$threshold", str(threshold))
            # this will do nothing if there's no subthreshold keyword in the message. All values are
            # fractions, and are expected to be percentages in the output
            if "subthreshold" in warning:
                thewarning = thewarning.replace("$subthreshold", f"{warning['subthreshold']*100:.0f}")
        if need_warning:
            # Warnings can be different in different subreports. This logic works because
            # subreports are processed first - if they exist, the main (combined) report's
            # warning is not triggered.
            if sub_word is not None:
                thename = f"{sub_str}{thename}"
                thewarning = f"{sub_word}: {thewarning}"
            # So that we don't get duplicate warnings from the main report. Note that when
            # processing the subreports, this will check for 11thename, which should
            # definitely not exist.
            if f"1{thename}" not in warnings and f"2{thename}" not in warnings:
                warnings[thename] = thewarning


class InstrumentConfig(TelescopeConfig):

    """
    This is a trivial top-level class for now. It overrides TelescopeConfig._get_config() to get the configuration
    information from a different place, currently under $pandeia_refdata/<tel_name>/<inst_name>
    """

    def _get_config(self):
        """
        Read default source configuration from JSON

        Returns
        -------
        config: dict
            All desired class attributes should have defaults defined in the config file
        """
        # use this trick to key the configuration file name off the name of the instantiated subclass
        self.inst_name = self.__class__.__name__.lower()
        self.ref_dir = os.path.join(default_refdata_directory, self.telescope.tel_name, self.inst_name)
        config = read_json(os.path.join(self.ref_dir, "config.json"), raise_except=True)

        # Grab the warning lists, if they exists, in the config file.
        self.warning_list_input = []
        if 'warning_list_input' in config:
            self.warning_list_input = config['warning_list_input']
        if 'warning_list_output' in config:
            self.warning_list_output = config['warning_list_output']

        # pop the defaults entry out
        if "defaults" in config:
            defaults = config.pop("defaults")

            if self.mode is None and not self.webapp:
                # if no mode specified, then get default mode
                if "mode" in defaults:
                    self.mode = defaults.pop("mode")
                else:
                    message = "No mode or default mode specified."
                    raise DataConfigurationError(value=message)
            # make sure mode is valid and if so use it to key the values to put into config
            msg = "Invalid %s mode: %s" % (self.inst_name, self.mode)
            config.update(self._load_dict(defaults[self.mode], "required", msg=msg))
            # we occasionally have required instrument and detector items we do NOT want to overwrite (JETC-3099)
            instrument_config = config.pop("instrument", {})
            detector = config.pop("detector", {})
            if self.webapp:
                config.pop("instrument_args", None)
            else:
                config.update(self._load_dict(defaults[self.mode], "instrument_args", msg=msg))
                # update and reinsert in the correct order
                instrument_config.update(config["instrument"])
                detector.update(config["detector"])
            config["detector"] = detector
            config["instrument"] = instrument_config

        return config


class Instrument(InstrumentConfig):

    """
    This is the primary Instrument class that sets up the configuration and provides methods common
    to all instruments.

    Parameters
    ----------
    telescope: Telescope sub-class instance
        The telescope that contains the instrument
    mode: str
        Desired instrument mode
    config: dict
        dictionary containing necessary configuration information
    **kwargs: list of keyword/value pairs
        parameter keyword and value pairs to augment defaults and config
    """

    def __init__(self, telescope=None, mode=None, config={}, webapp=False, **kwargs):
        self.webapp = webapp
        if telescope is None:
            msg = "Telescope not defined for %s!" % self.__class__.__name__
            raise EngineInputError(value=msg)
        if mode is None:
            msg = "Instrument mode not defined for %s!" % self.__class__.__name__
            raise EngineInputError(value=msg)
        self.telescope = telescope
        self.mode = mode
        self.order = None

        InstrumentConfig.__init__(self, config, webapp=webapp, **kwargs)

        # make sure telescope is configured properly. if not, the reference data directory
        # is likely not defined or doesn't exist.
        try:
            # make sure instrument is supported
            if self.inst_name not in self.telescope.instruments:
                message = "Unsupported %s instrument: %s" % (self.telescope.tel_name, self.inst_name)
                raise EngineInputError(value=message)
        except AttributeError as e:
            message = "Telescope/instrument configuration data missing.  "
            message += "Reference data directory possibly missing or not defined: %s" % default_refdata_directory
            message += "Original exception: %s" % e
            raise InternalError(value=message)

        # make sure required parameters are defined and do strict API checking if webapp is true
        all_config = merge_data(config, dict(**kwargs))
        self._nested_api_checks(all_config, webapp=webapp)

        self.read_detector()

        self.the_detector = self.get_detector_pars()

        # Check the instrument and strategy for the calculation against
        # the configured warnings in the instrument's config.json file.
        #
        # There are times that an Instrument instance will be created in order to just read
        # some values (e.g., bandpass in normalization.py's function _get_bandpass()) and in
        # these cases the only thing passed in is the instrument information. So, we are going
        # to check the warnings only if 'detector' and 'instrument' are passed in through all_config.
        if 'detector' in all_config and 'instrument' in all_config:

            # Each thing we want to be able to get the value of in the input dictionary
            # should be listed here. The "value" part is a list of nested keys to get the actual value.
            value_lookup = {
                'instrument': ['instrument', 'instrument'],
                'mode': ['instrument', 'mode'],
                'filter': ['instrument', 'filter'],
                'disperser': ['instrument', 'disperser'],
                'bin_spatial': ['configuration', 'detector', 'bin_spatial'],
                'bin_dispersion': ['configuration', 'detector', 'bin_dispersion'],
                'aperture': ['instrument', 'aperture'],
                'detector': ['instrument', 'detector'],
                'nexp': ['detector', 'nexp'],
                'ngroup': ['detector', 'ngroup'],
                'nint': ['detector', 'nint'],
                'postflash': ['detector', 'postflash'],
                'max_total_groups': ['detector', 'max_total_groups'],
                'readout_pattern': ['detector', 'readout_pattern'],
                'subarray': ['detector', 'subarray']
            }

            check_warnings(all_config, self.warnings, self.warning_list_input, value_lookup)

    def _nested_api_checks(self, conf, webapp=False):
        for section in self.instrument_pars:
            for p in self.instrument_pars[section]:
                # if parameter is not already set, then it's a DataError since config file should have set it to a default.
                if p not in getattr(self, section):
                    msg = "Missing required %s parameter %s from default configuration for %s." % (
                        section,
                        p,
                        self.__class__.__name__
                    )
                    raise DataError(value=msg)
                # if we're doing API checking, make sure everything we need is specified by the inputs.
                if webapp:
                    if p not in conf[section]:
                        key = "%s_missing_%s_%s" % (
                            self.__class__.__name__.lower(),
                            section,
                            p
                        )
                        self.warnings[key] = warning_messages["missing_instrument_api"].format(
                            section,
                            p,
                            self.__class__.__name__,
                            getattr(self, section)[p]
                        )

    def read_detector(self):
        """Read the detector out of the aperture data loaded from the config.json file and put the detector keyword
        in self.instrument['detector']
        This is only done by JWST and Roman, whose classes will override this function.  For HST, nothing will be done."""
        pass

    def _get_psf_key(self):
        """
        Get the PSF key for the current instrument configuration.  Generally, this is just the aperture, but other
        instruments use more complicated logic (COS, NIRCam coronagraphy, etc.).
        """
        return self.instrument['aperture']

    def _loadpsfs(self):
        """
        Load the PSF library for the current instrument configuration.  In most cases, this is based on the aperture.
        In some cases (e.g. NIRCam coronagraphy with bar-shaped masks), other configuration parameters come into play.
        """
        self.psf_library = self._load_psf_library(self._get_psf_key())

    def _load_psf_library(self,psf_key):

        psf_path = os.path.join(self.ref_dir, "psfs")
        library = PSFLibrary(self.get_wave_range(), path=psf_path, psf_key=psf_key)

        return library

    @property
    def qe_key(self):
        """
        set default qe_key for looking up QE reference data. MIRI different from others and overloads this.
        """
        return 'qe'

    def get_detector_pars(self):
        """
        Create the_detector (and Exposure_Spec) by passing in the appropriate detector_config, subarray_config, and readout_pattern_config options
        """

        # this det_dict has been parsed down to key: value pairs for a single detector; self.detector_config is the complete unparsed tree.
        det_dict = self.read_detector_pars() 

        detector_config = {"detector_config": det_dict, "subarray_config": self.subarray_config, 
            "readout_pattern_config": self.readout_pattern_config, "input_detector": self.detector}

        the_detector = DetectorFactory(config=detector_config, webapp=self.webapp)
        return the_detector

    def read_detector_pars(self):
        """
        Return detector-specific parameters.

        This reads the JSON structure so values can be dependent on aperture, subarray,
        readout pattern, and/or instrument mode.

        Returns
        -------
        det_dict: dict
            A dictionary of the correct detector parameters
        """
        det_dict = {}
        det_pars = self.detector_config[self.instrument['detector']]

        for parameter_key in det_pars:
            parameter_value = self.read_config_param(det_pars, parameter_key)

            det_dict[parameter_key] = parameter_value

        if 'excessp1' not in det_dict:
            det_dict['excessp1'] = 0.0
        if 'excessp2' not in det_dict:
            det_dict['excessp2'] = 0.0
        if 'rn_fudge' not in det_dict:
            det_dict['rn_fudge'] = 1.0
        if 'dark_current' not in det_dict:
            det_dict['dark_current'] = 0.0
        if 'rn' not in det_dict:
            det_dict['rn'] = 0.0
        if 'var_fudge' not in det_dict:
            det_dict['var_fudge'] = 1.0

        # if not explicitly set, assume we just need 2 groups to define a slope
        if 'mingroups' not in det_dict:
            det_dict['mingroups'] = 2

        # It was decided that only one of the fudge factors are allowed to be used, not both.
        # Issue #2261#issuecomment-260737726
        if det_dict['rn_fudge'] != 1.0 and (det_dict['excessp1'] != 0.0 or det_dict['excessp2'] != 0.0):
            raise ValueError('Only one of rn_fudge or excessp1/p2 maybe used, not both.')

        return det_dict

    def get_name(self):
        """
        Return instrument's name
        """
        return self.inst_name

    def get_aperture(self):
        """
        Return configured aperture
        """
        return self.instrument['aperture']

    def get_spatial_binning(self):
        """
        Return the spatial-dimension binning values. 
        
        Binning will be an integer or 1 if not present
        """

        if (("bin_spatial" in self.detector) and not self.the_detector.detector_config.get("has_binning", False)):
            raise EngineInputError(f"Detector does not support pixel binning.")
        bin_spatial = self.detector.get("bin_spatial", 1)
        if (not isinstance(bin_spatial, int)) or (bin_spatial < 1):
            raise EngineInputError(f"Spatial binning value {bin_spatial} invalid. Must be positive integer.")
        return bin_spatial

    def get_dispersion_binning(self):
        """
        Return the dispersion-direction binning values. 
        
        Binning will be an integer or 1 if not present
        """

        if (("bin_dispersion" in self.detector) and not self.the_detector.detector_config.get("has_binning", False)):
            raise EngineInputError(f"Detector does not support pixel binning.")
        bin_dispersion = self.detector.get("bin_dispersion", 1)
        if (not isinstance(bin_dispersion, int)) or (bin_dispersion < 1):
            raise EngineInputError(f"Dispersion binning value {bin_dispersion} invalid. Must be positive integer.")
        return bin_dispersion

    def get_binning(self):
        """
        Combined binning for the data
        """
        if self.projection_type in ["image", "image_scan"]:
            return self.detector.get("bin_spatial", 1) ** 2
        else:
            return self.detector.get("bin_spatial", 1) * self.detector.get("bin_dispersion", 1)

    def _interp_refdata(self, wave, key, colname, default=None):
        """
        Read in requested reference file and return requested data interpolated onto wave

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength array onto which the reference data is to be interpolated
        key: str
            Key to get filename from self.paths
        colname: str
            Column to retrieve from reference file
        default: None or float
            Default value to return in case of missing file or column

        Returns
        -------
        ref: numpy.ndarray or float
            If file exists, return reference_data(wave). Else return default.
        """
        ref = None
        if key is not None and key in self.paths:
            ref_file = os.path.join(self.ref_dir, self.paths[key])
            if ref_file is not None:
                ref = ref_data_interp(ref_file, wave, colname)
        if ref is None and default is None:
            msg = "No reference data found and no default value supplied for %s in %s." % (colname, key)
            raise DataError(value=msg)
        elif ref is None and default is not None:
            ref = np.ones_like(wave)*default
        return ref

    def _get_throughput(self, wave, key, colname='throughput'):
        """
        Wrap self._interp_refdata for getting throughput data
        HST does not and cannot use this method as-is because its high-resolution modes require 64-bit arrays
        
        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength array onto which the throughput curve is to be interpolated
        key: str
            Key to get filename from self.paths
        colname: str
            column name where throughput is stored in table

        Returns
        -------
        eff: numpy.ndarray or float
            If ref file exists, return efficiency(wave), else return 1.0
        """
        eff = self._interp_refdata(wave, key, colname=colname, default=1.0)
        return np.array(eff, dtype=np.float32)

    def _get_disperser(self):
        """
        The same disperser can have different configurations depending on which order is
        being used. Use self.order to build a new key for looking up configuration data for the
        given order.

        Returns
        -------
        key: str
            Key to look up order-specific configuration data
        """
        key = self.instrument['disperser']
        if self.order is not None:
            key = "%s_%d" % (key, self.order)
            key = key.replace('-','m')
        return key

    def _get_dispersion_key(self):
        """
        Different instruments have different ways to figure out which disperser to use. This is also used
        to look up reference data for multiple orders from the same disperser.

        Returns
        -------
        key: str
            Key used to get dispersion file out of self.paths
        """
        disperser = self._get_disperser()
        key = "%s_disp" % disperser
        return key

    def get_filter_eff(self, wave):
        """
        Read in filter throughput curve

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto

        Returns
        -------
        eff: numpy.ndarray or float
            Filter throughput as a function of wave
        """
        eff = self._get_throughput(wave, self.instrument['filter'])
        return eff

    def get_disperser_eff(self, wave):
        """
        Read in efficiency curve for configured grating

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto

        Returns
        -------
        eff: numpy.ndarray or float
            Disperser efficiency as a function of wave
        """
        key = self._get_disperser()
        eff = self._get_throughput(wave, key)
        return eff

    def get_internal_eff(self, wave, full_throughput=False):
        """
        Read in internal efficiency

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto
        full_throughput: bool
            Controls whether certain instrument-specific variants use special behavior.

        Returns
        -------
        eff: numpy.ndarray or float
            Internal throughput as a function of wave
        """
        eff = self._get_throughput(wave, 'internal')
        return eff

    def get_detector_qe(self, wave):
        """
        Read in detector quantum efficiency. This will likely need to be
        overloaded on a per detector basis.

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto

        Returns
        -------
        qe: numpy.ndarray or float
            Quantum efficiency as a function of wave
        """
        qe = self._get_throughput(wave, self.qe_key)
        return qe

    def get_quantum_yield(self, wave):
        """
        Read in detector quantum yield if it's there. If not, return unity (trivial yield).

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto

        Returns
        -------
        quantum_yield: numpy.ndarray or float
            Quantum efficiency as a function of wave
        """
        q_yield = self._interp_refdata(wave, self.qe_key, colname='conversion', default=1.0)
        fano_factor = (3.0 * q_yield - q_yield**2. - 2.) / q_yield  # two level model, valid for quantum yield between 1 and ~2.
        return q_yield, fano_factor

    def apply_scattering(self, extracted):
        """
        Compute additional scattered light term

        This function is overridden when used.
        """

        return extracted

    def get_global_scattering(self):
        """
        Get the global scattering term

        This function is overridden when used.
        """

        return 0.0

    def get_total_eff(self, wave, full_throughput=False):
        """
        Calculate combined throughput of entire telescope/instrument/detector system

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate throughput onto

        Returns
        -------
        eff: numpy.ndarray or float
            Total system throughput as a function of wave
        """
        ote_int = self.telescope.get_ote_eff(wave)
        filter_eff = self.get_filter_eff(wave)
        disperser_eff = self.get_disperser_eff(wave)
        internal_eff = self.get_internal_eff(wave, full_throughput=full_throughput)
        det_qe = self.get_detector_qe(wave)

        eff = ote_int * filter_eff * disperser_eff * internal_eff * det_qe
        return eff

    def get_dispersion(self, wave):
        """
        Read in dispersion

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate/trim dispersion data onto

        Returns
        -------
        dispersion: numpy.ndarray or float
            Dispersion as a function of wave
        """
        disperser = self._get_disperser()
        if disperser is not None:
            key = self._get_dispersion_key()
            dispersion = self._interp_refdata(wave, key, colname='dlds', default=DEFAULT_DISPERSION)
        else:
            dispersion = 0.0
        return dispersion

    def get_trace(self, wave):
        """
        Implement spectral traces, wavelength-dependent shifts of the spatial axis.  This base implementation only returns
        an array of 0's of the appropriate size.  This is to make the code implementation of this fully general.
        Instruments that need to account for such spatial distortions will need to overload this to load and implement
        the appropriate reference data.

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate/trim trace data onto

        Returns
        -------
        trace: numpy.ndarray or float
            Spectral trace offset from center of FOV as a function of wave
        """
        trace = np.zeros_like(wave)

        return trace

    def get_detector_pixels(self, wave):
        """
        Read in detector pixel positions for each wavelength in wave_pix

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate/trim pixel position data onto

        Returns
        -------
        pixels: numpy.ndarray or None
            Detector pixel positions corresponding to each element of wave
        """
        disperser = self._get_disperser()
        if disperser is not None:
            key = "%s_wavepix" % disperser
            try:
                pixels = self._interp_refdata(wave, key, colname='pixels', default=np.arange(len(wave)))
            except DataError as e:
                pixels = None
        else:
            pixels = None
        return pixels

    def get_resolving_power(self, wave):
        """
        Read in resolving power

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to interpolate resolving power onto

        Returns
        -------
        rpower: numpy.ndarray or None
            Resolving power as a function of wave
        """
        disperser = self._get_disperser()
        if disperser is not None:
            key = self._get_dispersion_key()
            rpower = self._interp_refdata(wave, key, colname='R', default=wave/(DEFAULT_DISPERSION*2))
        else:
            rpower = None
        return rpower

    def get_wave_blaze(self):
        """
        Get wavelength vector used in the grating efficiency (blaze) file

        Returns
        -------
        wave_blaze: numpy.ndarray
            Wavelength vector from the grating efficiency file
        """
        key = self._get_disperser()
        blaze_file = os.path.join(self.ref_dir, self.paths[key])
        msg = "Blaze file not found: %s" % blaze_file
        wave_blaze = ref_data_column(blaze_file, colname='wavelength', error_msg=msg)
        return wave_blaze

    def get_wave_filter(self):
        """
        Get wavelength vector used in the filter throughput file

        Returns
        -------
        wave_filter: numpy.ndarray
            Wavelength vector from the filter throughput file
        """
        key = self.instrument['filter']
        filter_file = os.path.join(self.ref_dir, self.paths[key])
        msg = "Filter file not found: %s" % filter_file
        wave_filter = ref_data_column(filter_file, colname='wavelength', error_msg=msg)
        return wave_filter

    def get_wave_pix(self):
        """
        Get wavelength vector to convert pixel position to wavelength

        Returns
        -------
        wavepix: numpy.ndarray
            Wavelength vector mapping pixels to wavelengths
        """
        # usually wavepix is lifted from the dispersion file, but in some cases (e.g. NIRISS SOSS)
        # there is a specific, calibrated mapping of wavelength to pixels. look for that first and then
        # fall back to the dispersion file.
        wavepix_key = "%s_wavepix" % self._get_disperser()
        if wavepix_key in self.paths:
            # if we have a specifically calibrated mapping of wavelengths to pixels (e.g. for SOSS),
            # look up that file and use that mapping.
            wavecal_file = os.path.join(self.ref_dir, self.paths[wavepix_key])
            msg = "Wavelength calibration file not found: %s" % wavecal_file
            wavepix = ref_data_column(wavecal_file, colname='wavelength', error_msg=msg)
        else:
            # in dispersion files, there is no guarantee that the table is sampled in pixel spacing. This information
            # is encoded in the dlds column so calculate it from there.
            key = self._get_dispersion_key()
            if key in self.paths:
                dispersion_file = os.path.join(self.ref_dir, self.paths[key])
                msg = "Dispersion file not found: %s" % dispersion_file
                wave = ref_data_column(dispersion_file, colname='wavelength', error_msg=msg)
                dlds = ref_data_column(dispersion_file, colname='dlds', error_msg=msg)
            else:
                # for spectroscopic order 0, there will be no dispersion file, and thus no reference in the
                # config.json path{} block.
                wave = self.get_wave_blaze()
                dlds = np.ones_like(wave) * DEFAULT_DISPERSION
            pixel = np.cumsum(1.0 / dlds * np.gradient(wave))
            pixel_integer = np.arange(int(pixel[0]), int(pixel[-1]))
            wavepix = np.interp(pixel_integer, pixel, wave)

        return wavepix

    def get_chip_dimensions(self, wave):
        """
        This function allows us to specify the dimensions of a chip or area on a detector
        that have their own properties.
        
        Use case: 
        COS FUV detector is comprised of two chips that have their own
        dark current, and require their own health and safety outputs; the spectrum 
        is produced as a single array with a gap. COS will need to override this function
        to produce that output.
        (NIRSpec does the same thing but with identical detectors that do not have
        health and safety concerns)

        Thus, we have both the complete dimensions of the chip (in pixel space) and the 
        portion of the computed spectrum on that chip to consider.

        """
        nx = self.the_detector.pixels_x
        ny = self.the_detector.pixels_y

        dimensions = [{"name": self.instrument.get("detector","default"), 
                        "size":{"x": nx, "y": ny}, "bounds": {"x": slice(-1), "y": slice(-1)}}]
        return dimensions

    def get_ipc_kernel(self):
        """
        Get inter-pixel capacitance (IPC) kernel

        Returns
        -------
        kernel: numpy.ndarray
            IPC kernel data
        """
        key = "ipc_kernel"
        ipc_file = os.path.join(self.ref_dir, self.paths[key])
        try:
            kernel = fits.getdata(ipc_file)
        except IOError as e:
            msg = "Error reading IPC kernel reference file: %s. " % ipc_file
            if self.webapp:
                msg += "(%s)" % type(e)
            else:
                msg += repr(e)
            raise DataError(value=msg)
        # The new convolution algorithm in scipy 1.4+ can't handle comparing data of different types
        return np.asarray(kernel,dtype=np.float32)

    def get_variance_fudge(self, wave):
        """
        Some instruments (e.g. MIRI) have measured noise properties that are not well explained
        or modeled by Pandeia's current noise model. This method returns the multiplicative fudge factor
        that is to be applied to the pixel rate variance to match IDT expectations. The fudge factor can be
        chromatic so needs to accept an array of wavelengths to interpolate over. By default this returns
        a 1's array the same length as wave scaled by the 'var_fudge' scalar value defined in detector_config.

        Parameters
        ----------
        wave: 1D numpy.ndarray
            Wavelengths to interpolate chromatic variance over
        """
        return self.the_detector._get_variance_fudge(wave)

    def get_readnoise_correlation_matrix(self, shape):
        """
        Grab correlated readnoise data out of reference file and build a correlation
        matrix out of it.

        Parameters
        ----------
        shape: list-like

        Returns
        -------
        correlation_matrix: Scipy.sparse.csr_matrix
            Compressed sparse row correlation matrix
        """
        key = "rn_corr"
        correlation_file = os.path.join(self.ref_dir, self.paths[key])
        try:
            correlation = fits.getdata(correlation_file)
        except IOError as e:
            msg = "Error reading RN correlation reference file: %s. " % correlation_file
            if self.webapp:
                msg += "(%s)" % type(e)
            else:
                msg += repr(e)
            raise DataError(value=msg)

        nframe_range = correlation.shape[0]
        nx = correlation.shape[2]
        ny = correlation.shape[1]

        if nx != ny:
            message = 'Only square correlation matrices are supported.'
            raise UnsupportedError(value=message)

        if self.the_detector.exposure_spec.nframe > nframe_range:
            # nframe longer than full range of correlation so use the full range
            corr_range = nframe_range
        else:
            # nframe shorter than range of correlation so use subset of correlation data
            corr_range = self.the_detector.exposure_spec.nframe
        correlation_data = sparse.csr_matrix(correlation[corr_range, :, :])

        empty_matrix = sparse.coo_matrix((1024, 1024))
        correlation_matrix = sparse.bmat([[empty_matrix, None, empty_matrix],
                                          [None, correlation_data, None],
                                          [empty_matrix, None, empty_matrix]])
        correlation_matrix = sparse.csr_matrix(correlation_matrix)

        return correlation_matrix

    def read_config_param(self, item, parameter_key):
        """
        Some instrument parameters are dependent on the specifics of detector or instrument setup. This is a generic
        function to pull the correct values out of a dictionary block.
        This is meant to work on dictionaries that are not themselves broken into different filters, gratings, or
        readout patterns; it's for what would otherwise be a flat dictionary of configuration values, like detector_config.
        Values that depend on multiple parameters can be used; the key_format list will contain both parameters and
        the options will contain both relevant values concatenated by "__"
        An example of what a dictionary to be parsed would look like:
        "default": {
            "det_type": "h2rg",
            "dark_current": 0.00636,
            "excessp1": 0.0,
            "excessp2": 0.0,
            "ff_electrons": 62500,
            "fullwell": {
    			"default": 65000.0,
    			"key_format": ["subarray", "readout_pattern"],
    			"full__nrsrapid": 30000.0,
    			"s200a1__nrs": 22000.0
    		},
            "ipc": true,
            "mingroups": 2,
            "pix_size": 18.0,
            "rn": {
    			"default": 16.8,
    			"key_format": ["aperture"],
    			"sw": 17
    		},
            "rn_correlation": true
        }
        item: dict
            Dictionary within which we need to search for the correct detector configuration parameter.
        parameter_key: string
            The parameter we want to search for
        returns: string, int, float, or bool
             The value of parameter_key for the given instrument and detector setup.
        """

        value_lookup = {
            "detector": self.instrument["detector"],
            "aperture": self.instrument["aperture"],
            "disperser": self.instrument["disperser"],
            "cenwave": self.instrument.get("cenwave", None),
            "filter": self.instrument["filter"],            
            "instmode": self.instrument["mode"],
            "readout_pattern": self.detector.get("readout_pattern", None),
            "gain": "a{}".format(self.detector.get("gain", None)),
            "binning": "b{}".format(self.get_binning()),
            "subarray": self.detector.get("subarray", None),
            "ngroup": "g{}".format(self.detector.get("ngroup", None)),
            "dark_level": self.detector.get("dark_level", None),
        }

        if parameter_key in item:
            if not isinstance(item[parameter_key], dict):
                # If the field exists and is not a dict then just return the value
                parameter_value = item[parameter_key]

            elif isinstance(item[parameter_key], dict) and "key_format" in item[parameter_key]:
                # If the field exists and it is a dict then we have to check the key_format
                # In this case the format of the parameter will be something like:
                # "rn": {
                #     "default": 16.8,
                #     "key_format": ["aperture"],
                #     "sw": 17
                # },

                # start with the default; if nothing else applies we'll return that value
                if "default" in item[parameter_key]:
                    parameter_value = item[parameter_key]['default']
                else:
                    raise DataError("There is no default value in {}".format(item[parameter_key]))
                # Create the key lookup
                key_lookup = []
                for kf in item[parameter_key]["key_format"]:
                    kfelement = kf.split("__")
                    for kfe in kfelement:
                        try:
                            key_lookup.append(str(value_lookup[kfe]))
                            # if this value alone is in the item listing, grab it
                            # (if more than one shows up, there has to be a listing with both)
                            if str(value_lookup[kfe]) in item[parameter_key]:
                                parameter_value = item[parameter_key][str(value_lookup[kfe])]
                        except KeyError as e:
                            raise DataError(
                                "Unknown key format {} value {} listed in {}".format(kfe, value_lookup[kfe],
                                                                                     item[parameter_key]))

                # Now squish the list to a "__" separated key and see if it exists
                # If all required keys were passed in, then create the string lookup.
                # If the string is present in the dict, return it.
                if not any([x is None for x in key_lookup]):
                    key_lookup = "__".join(key_lookup)

                    if key_lookup in item[parameter_key]:
                        parameter_value = item[parameter_key][key_lookup]

            else:
                raise DataError("Improperly formatted parameter {}".format(item))
        else:
            raise DataError("There is no {} key in {}".format(parameter_key, item))

        return parameter_value

    def slitlet_grid(self,slitlet):
        nx = self.slitlet_config[slitlet]["nx"]
        ny = self.slitlet_config[slitlet]["ny"]
        xstep = self.slitlet_config[slitlet]["xstep"]
        ystep = self.slitlet_config[slitlet]["ystep"]
        positions = []
        startx = -(nx - 1) / 2.0 * xstep
        stopx = (nx - 1) / 2.0 * xstep
        starty = -(ny - 1) / 2.0 * ystep
        stopy = (ny - 1) / 2.0 * ystep
        startx = int(startx) if startx % 1 == 0 else int(startx - 0.5)
        stopx = int(stopx) if stopx % 1 == 0 else int(stopx - 0.5)
        starty = int(starty) if starty % 1 == 0 else int(starty - 0.5)
        stopy = int(stopy) if stopy % 1 == 0 else int(stopy - 0.5)
        for x in range(startx,stopx+1,xstep):
            for y in range(starty,stopy+1,ystep):
                positions.append([x,y])
        return positions

    def get_slit_pars(self):
        """
        Collect instrument slit parameters into a dict and return it.
        This assumes if a slit is defined in the input dictionary, there
        will be a slit defined in the config.json data.

        Returns
        -------
        slit_dict: dict
            Contains the following parameters describing the instrument aperture:
                disp - width of slit in dispersion direction (X-axis)
                xdisp - height of slit in cross-dispersion direction (Y-axis)
                disp_pitch - distance between adjacent slits in the dispersion direction
                             (None unless a multi-shutter instrument)
                xdisp_pitch - distance between adjacent slits in the cross-dispersion direction
                              (None unless a multi-shutter instrument)
                nslice - number of slices the aperture is divided into (for IFUs)
                slit_shape - "rectangle" or "circle"; defaults to "rectangle"
        """
        if "slit" in self.instrument:
            slit = self.instrument["slit"]
            try:
                slit_config = self.slit_config[slit]
            except KeyError:
                raise DataError("Slit {} not recognized".format(slit))
            disp = slit_config["disp"]
            xdisp = slit_config["xdisp"]
            nslice = slit_config["nslice"]
            slit_shape = slit_config.get("shape","rectangle")
            slit_axis = slit_config.get("axis", "x")

            """
            For multi-shutter instruments, the shutters may have gaps between them, represented 
            as a shutter size smaller than the offset between the individual shutters (the pitch). 
            If so, the pitch should be provided in the instrument configuration.
            """
            if "disp_pitch" in slit_config:
                try:
                    disp_pitch = slit_config["disp_pitch"]
                    xdisp_pitch = slit_config["xdisp_pitch"]
                except KeyError as e:
                    raise DataError("Malformed config file: missing entry for shutter pitch (%s)" % e)
            else:
                """
                If the pitch is not on the config, assume it is the same as the shutter size.
                """
                disp_pitch = disp
                xdisp_pitch = xdisp

            if "slitlet_shape" in self.instrument:
                if isinstance(self.instrument["slitlet_shape"], str):
                    slitlet_shape = self.slitlet_grid(self.instrument["slitlet_shape"])
                else:
                    slitlet_shape = self.instrument["slitlet_shape"]
                """
                The shutter pattern user input is in units of shutter pitch, so convert to internal angular units here.
                """
                multishutter = [(disp_pitch * shutter[0], xdisp_pitch * shutter[1]) for shutter in slitlet_shape]
            else:
                multishutter = None

        else:
            disp = None
            xdisp = None
            nslice = None
            multishutter = None
            slit_shape = None
            slit_axis = None

        slit_dict = {
            "disp": disp,
            "xdisp": xdisp,
            "multishutter": multishutter,
            "nslice": nslice,
            "slit_shape": slit_shape,
            "slit_axis": slit_axis
        }
        return slit_dict

    def get_aperture_pars(self):
        """
        Collect instrument aperture parameters into a dict and return it

        Returns
        -------
        aperture_dict: dict
            Contains the following parameters describing the instrument aperture:
                pix - pixel scale in arcsec/pixel
                detector - which detector it goes with
                plate_scale - pixel scale, y and x dimensions. Should match PSFs. 
                    Defaults to match (pix, pix) if not specified, which may happen when pixels
                    are truly square.
        """
        aperture = self.get_aperture()
        pix = self.aperture_config[aperture]["pix"]
        plate_scale = (self.aperture_config[aperture].get("plate_scale_y", pix), self.aperture_config[aperture].get("plate_scale_x", pix))

        aperture_dict = {
            "pix": pix,
            "plate_scale": plate_scale
        }
        return aperture_dict

    def get_wave_range(self):
        """
        Get the wavelength range of the current instrument configuration. The wavelength
        range information is organized in the configuration files on a per aperture basis.

        Returns
        -------
        range_dict: dict
            Contains the instrument wavelength range in microns described by:
                wmin - minimum wavelength
                wmax - maximum wavelength
        """
        disperser = self._get_disperser()
        aperture = self.instrument['aperture']
        filt = self.instrument['filter']
        cenwave = self.instrument.get("cenwave", None)
        slit = self.instrument.get("slit", None)

        # the patterns supported are:
        # for multiorder:
        #   aperture - disperser - filter
        #   aperture - disperser - cenwave
        #   aperture - disperser
        # for other projections:
        #    aperture - disperser - filter
        #    aperture - disperser - cenwave
        #    aperture - disperser
        #    aperture - filter
        #    aperture - slit (where slit is probably a filtered slit)
        #    aperture

        # if disperser is defined and there are range entries for it, it takes precendence
        # For multiorder projections, we need to scan through a number of dispersion orders, named disperser_n in the
        # ranges section of the configuration file:
        if self.projection_type in "multiorder":
            wmins = []
            wmaxes = []
            for disp in self.range[aperture]:
                # make sure this is one of the disperser's ranges.
                if disperser in disp:
                    # the order may go disperser - filter - cenwave or disperser - cenwave
                    if filt is not None and filt in self.range[aperture][disp]:
                        wmins.append(self.range[aperture][disp][filt]["wmin"])
                        wmaxes.append(self.range[aperture][disp][filt]["wmax"])
                    elif cenwave is not None and cenwave in self.range[aperture][disp]:
                        wmins.append(self.range[aperture][disp][cenwave]["wmin"])
                        wmaxes.append(self.range[aperture][disp][cenwave]["wmax"])
                    else:
                        wmins.append(self.range[aperture][disp]["wmin"])
                        wmaxes.append(self.range[aperture][disp]["wmax"])
            # get the min of mins and max of maxes
            wmin = min(wmins)
            wmax = max(wmaxes)
        elif disperser is not None and disperser in self.range[aperture]:
            # the order may go disperser - filter - cenwave or disperser - cenwave
            if filt is not None and filt in self.range[aperture][disperser]:
                wmin = self.range[aperture][disperser][filt]["wmin"]
                wmax = self.range[aperture][disperser][filt]["wmax"]
            elif cenwave is not None and cenwave in self.range[aperture][disperser]:
                wmin = self.range[aperture][disperser][cenwave]["wmin"]
                wmax = self.range[aperture][disperser][cenwave]["wmax"]
            else:
                wmin = self.range[aperture][disperser]["wmin"]
                wmax = self.range[aperture][disperser]["wmax"]
        # if filter is defined and has range entries, use it next
        elif filt is not None and filt in self.range[aperture]:
            # Possibility for dependency of the filter on the wavelength range.
            wmin = self.range[aperture][filt]["wmin"]
            wmax = self.range[aperture][filt]["wmax"]
        elif slit is not None and slit in self.range[aperture]:
            wmin = self.range[aperture][slit]["wmin"]
            wmax = self.range[aperture][slit]["wmax"]
        # otherwise fall back to the generic range for the current aperture
        else:
            wmin = self.range[aperture]["wmin"]
            wmax = self.range[aperture]["wmax"]

        # now check wmin and wmax against wave_pix for sanity, if applicable
        if self.projection_type in ('slitless', 'spec', 'slitless_scan'):
            wave_pix = self.get_wave_pix()
            wmin = max(wmin, wave_pix.min())
            wmax = min(wmax, wave_pix.max())

        range_dict = {"wmin": wmin, "wmax": wmax}
        return range_dict

    @property
    def projection_type(self):
        """
        Determine the appropriate projection type based on the configured instrument mode

        Returns
        -------
        proj_type: str
            The projection type, currently one of 'slitless', 'slitless_scan', 'spec', 'image', or 'image_scan'.
        """
        mode = self.mode
        slitless_modes = self.telescope.slitless_modes
        spec_modes = self.telescope.spec_modes
        image_modes = self.telescope.image_modes
        if hasattr(self.telescope, 'multiorder_modes'):
            multiorder_modes = self.telescope.multiorder_modes
        else:
            multiorder_modes = []
        if hasattr(self.telescope, 'slitless_scan_modes'):
            slitless_scan_modes = self.telescope.slitless_scan_modes
        else:
            slitless_scan_modes = []
        if hasattr(self.telescope, 'image_scan_modes'):
            image_scan_modes = self.telescope.image_scan_modes
        else:
            image_scan_modes = []

        if mode in slitless_modes:
            proj_type = 'slitless'
        elif mode in spec_modes:
            proj_type = 'spec'
        elif mode in image_modes:
            proj_type = 'image'
        elif mode in multiorder_modes:
            proj_type = 'multiorder'
        elif mode in slitless_scan_modes:
            proj_type = 'slitless_scan'
        elif mode in image_scan_modes:
            proj_type = 'image_scan'
        else:
            raise DataError("Unrecognized projection type for mode {}".format(mode))

        return proj_type

    def dispersion_axis(self):
        """
        The dispersion axis is either along rows (the X axis) or along columns (Y axis).  By default
        it is along the X axis.  Instrument subclasses will overload this method to deal with cases
        where it is along the Y axis (e.g. NIRISS GR150C).

        Returns
        -------
        disp_axis: str
            Axis along with spectra are dispersed.  Allowed values are 'x' or 'y'.
        """
        disp_axis = "x"
        return disp_axis

    def get_pix_size(self):
        """
        Shortcut to get the size of the pixels in the current observing mode/aperture combination.

        Returns
        -------
        pix_size: tuple
            The pixel scale in the y and x directions in arcsec/pixel
        """
        aperture_dict = self.get_aperture_pars()
        pix_size = aperture_dict['plate_scale']

        return pix_size

    def get_exposure_time(self):
        """
        Shortcut to get the total on-source time for the current detector configuration

        Returns
        -------
        time: float
            Total on-source time in seconds
        """
        time = self.the_detector.exposure_spec.exposure_time
        return time

    def create_gap_mask(self, wave):
        """
        Use the gap configuration and a wavelength vector, wave, to build a masked array that
        masks out the location of the gap.  Wavelengths between and including both gap endpoints
        will be masked out.

        For now this is a stub and instrument subclasses are expected to overload this to provide
        a gap mask.  Currently only NIRSpec does this.

        Parameters
        ----------
        wave: numpy.ndarray
            Wavelength vector to construct mask from

        Returns
        -------
        mask: float (default: 1.0)
            1D array masked at the locations within the configured detector gap and 1.0 elsewhere
        """
        return 1.0

    def spectrometer_convolve(self, spec):
        """
        This method convolves a spectrum with a wavelength-dependent spectral
        resolving power R=lambda/delta lambda, where delta lambda is the FWHM of an
        unresolved line at a wavelength lambda. This is accomplished by internally
        changing the sampling of the spectrum, such that delta-lambda is sampled by
        exactly a constant number of samples as a function of wavelength. The convolution
        is then accomplished using the convolution theorem and knowledge that the kernel is
        independent of wavelength.

        Parameters
        ----------
        spec: pandeia.engine.astro_spectrum.AstroSpectrum instance
            The input spectrum to convolve

        Returns
        -------
        spec: pandeia.engine.astro_spectrum.AstroSpectrum instance
            Same dimensions as input spec, but now convolved according to the wavelength-dependent resolving
            power of the instrument, if applicable.  Same as the input spec in the case of no dispersion (i.e. imaging).
        """
        r = self.get_resolving_power(spec.wave)
        if r is None:
            # no dispersion defined so return unmolested spectrum
            return spec

        # find the minimum spacing between wavelengths
        dws = np.abs(spec.wave - np.roll(spec.wave, 1))
        dw_min = np.min(dws)
        dw_arg = np.argmin(dws)

        # check that the minimum spacing is not too small, as this can lead to huge arrays. We will not
        # support sampling better than 2e5 (1e5 is the highest current resolving power of any IR spectrometer).
        if (spec.wave[dw_arg]/dw_min>self.telescope.minimal_spectral_spacing):
            dw_min = spec.wave[dw_arg]/self.telescope.minimal_spectral_spacing
            key = "spectral_resampling_too_small"
            self.warnings[key] = warning_messages[key].format(
                self.inst_name,
                spec.wave[dw_arg] / dw_min,
                self.telescope.minimal_spectral_spacing
            )

        fwhm = spec.wave / r  # FWHM of resolution element as a function of wavelength
        fwhm_s = np.min(fwhm / dw_min)  # find min of FWHM expressed in units of minimum spacing (i.e. sampling)

        # but do not allow the sampling FWHM to be less than Nyquist
        fwhm_s = np.max([2., fwhm_s])

        # divide FWHM(wave) by the FWHM of the sampling to get sampling as a function of wavelength
        ds = fwhm / fwhm_s

        # use the min as a starting point; initialize array
        w = np.min(spec.wave)

        # it's much faster (~50%) to append to lists than np.array()'s
        wave_constfwhm = []

        # doing this as a loop is slow, but straightforward.
        while w < np.max(spec.wave):
            # use interpolation to get delta-wavelength from the sampling as a function of wavelength.
            # this method is over 5x faster than the original use of scipy.interpolate.interp1d.
            w += np.interp(w, spec.wave, ds)
            wave_constfwhm.append(w)

        wave_constfwhm.pop()  # remove last point which is an extrapolation
        wave_constfwhm = np.array(wave_constfwhm)

        # interpolate the flux onto the new wavelength set
        flux_constfwhm = spectrum_resample(spec.flux, spec.wave, wave_constfwhm)

        # convolve the flux with a gaussian kernel; first convert the FWHM to sigma
        sigma_s = fwhm_s / 2.3548
        g = Gaussian1DKernel(sigma_s)

        # use boundary='extend' to set values outside the array to nearest array value.
        # this is the best approximation in this case.
        flux_conv = convolve(flux_constfwhm, g, normalize_kernel=True, boundary='extend')
        flux_oldsampling = np.interp(spec.wave, wave_constfwhm, flux_conv)

        spec.flux = flux_oldsampling
        return spec

    def get_scan_length(self):
        """
        Get, or compute, the length of a scan for a scanned spec or scan image mode.
        Users always have to specify the exposure_time (or components sufficient to calculate it); 
        if they scan, they must provide either scan_length (arcseconds) or scan_rate (arcseconds/sec), 
        otherwise this code raises an EngineInputError.

        If users specify BOTH a scan_length AND scan_rate, the code raises an EngineInputError if 
        the values are inconsistent.

        Returns
        -------
        scan_length: float
            length of the scan in arcseconds

        """
        if "scan_length" in self.the_detector.input_detector:
            scan_length = self.the_detector.input_detector["scan_length"]
        elif "scan_rate" in self.the_detector.input_detector:
            scan_length = self.the_detector.input_detector["scan_rate"] * self.the_detector.exposure_spec.exposure_time
        else:
            raise EngineInputError("Cannot define scan without scan length or scan rate.")
        # And now a sanity check: If all three are defined, are they consistent?
        if "scan_rate" in self.the_detector.input_detector:
            # This will be trivially true if we just computed the scan length from the scan rate.
            if scan_length != (self.the_detector.input_detector["scan_rate"] * self.the_detector.exposure_spec.exposure_time):
                raise EngineInputError("Scan definition inconsistent: time {}, rate {}, length {}".format(self.the_detector.exposure_spec.exposure_time, self.the_detector.input_detector["scan_rate"], scan_length))
        return scan_length

    def get_segment_indicies(self, wave):
        """
        Generic code to get the boundaries of a chip/stripe/segment in pixel coordinates, from our segment
        definitions. Functions that need this will override it.

        Parameters
        ----------
        wave: np.ndarray
            The wavelength array of the projected grid

        Returns
        -------
        segment_indices: list
            A list of dictionaries defining the segments in pixel coordinates.
        """

        segment_indicies = [{"segname": "", "wmin_ind": 0, "wmax_ind": len(wave)}]
        
        return segment_indicies
